from collections import OrderedDict

from .. import Provider as ColorProvider

localized = True


class Provider(ColorProvider):
    all_colors = OrderedDict((
        ("Açafrão", "#F4C430"),
        ("Água-marinha média", "#66CDAA"),
        ("Água-marinha", "#7FFFD4"),
        ("Água", "#00FFFF"),
        ("Alizarina", "#E32636"),
        ("Amarelo brasilis", "#ECDB00"),
        ("Amarelo claro", "#FFFFE0"),
        ("Amarelo creme", "#ECD690"),
        ("Amarelo escuro", "#F2B73F"),
        ("Amarelo esverdeado", "#9ACD32"),
        ("Amarelo esverdeado", "#ADFF2F"),
        ("Amarelo ouro claro", "#FAFAD2"),
        ("Amarelo queimado", "#EEAD2D"),
        ("Amarelo", "#FFFF00"),
        ("Âmbar", "#FFBF00"),
        ("Ameixa", "#DDA0DD"),
        ("Amêndoa", "#FFEBCD"),
        ("Ametista", "#9966CC"),
        ("Aspargo", "#7BA05B"),
        ("Azul aço claro", "#B0C4DE"),
        ("Azul aço", "#4682B4"),
        ("Azul alice", "#F0F8FF"),
        ("Azul ardósia claro", "#8470FF"),
        ("Azul ardósia escuro", "#483D8B"),
        ("Azul ardósia médio", "#7B68EE"),
        ("Azul ardósia", "#6A5ACD"),
        ("Azul areado", "#B8CAD4"),
        ("Azul brasilis brilhante", "#09ACDB"),
        ("Azul brasilis", "#00BDCE"),
        ("Azul cadete", "#5F9EA0"),
        ("Azul camarada", "#054F77"),
        ("Azul celeste brilhante", "#007FFF"),
        ("Azul celeste pernambucano", "#00A4CD"),
        ("Azul celeste", "#F0FFFF"),
        ("Azul céu claro", "#87CEFA"),
        ("Azul céu profundo", "#00BFFF"),
        ("Azul céu", "#87CEEB"),
        ("Azul claro", "#ADD8E6"),
        ("Azul cobalto", "#0047AB"),
        ("Azul escuro", "#00008B"),
        ("Azul flor de milho", "#6495ED"),
        ("Azul força aérea", "#5D8AA8"),
        ("Azul furtivo", "#1E90FF"),
        ("Azul manteiga", "#A6AA3E"),
        ("Azul marinho", "#120A8F"),
        ("Azul médio", "#0000CD"),
        ("Azul meia-noite", "#191970"),
        ("Azul petróleo", "#084D6E"),
        ("Azul pólvora", "#B0E0E6"),
        ("Azul real", "#0000DD"),
        ("Azul taparuere", "#248EFF"),
        ("Azul turquesa brilhante", "#00DDFF"),
        ("Azul turquesa", "#00CCEE"),
        ("Azul violeta", "#8A2BE2"),
        ("Azul", "#0000FF"),
        ("Bege", "#F5F5DC"),
        ("Bordô", "#800000"),
        ("Borgonha", "#900020"),
        ("Branco antigo", "#FAEBD7"),
        ("Branco fantasma", "#F8F8FF"),
        ("Branco floral", "#FFFAF0"),
        ("Branco fumaça", "#F5F5F5"),
        ("Branco navajo", "#FFDEAD"),
        ("Branco", "#FFFFFF"),
        ("Brasil", "#A7F432"),
        ("Bronze", "#CD7F32"),
        ("Caqui escuro", "#BDB76B"),
        ("Caqui", "#F0E68C"),
        ("Caramelo", "#8B5742"),
        ("Cardo", "#D8BFD8"),
        ("Carmesim", "#DC143C"),
        ("Carmim carnáceo", "#960018"),
        ("Carmim clássico", "#992244"),
        ("Carmim", "#712F26"),
        ("Castanho avermelhado", "#8B0000"),
        ("Castanho claro", "#D2B48C"),
        ("Cenoura", "#ED9121"),
        ("Cereja Hollywood", "#F400A1"),
        ("Cereja", "#DE3163"),
        ("Chocolate", "#D2691E"),
        ("Ciano claro", "#E0FFFF"),
        ("Ciano escuro", "#008B8B"),
        ("Ciano", "#00FFFF"),
        ("Cinza ardósia claro", "#778899"),
        ("Cinza ardósia escuro", "#2F4F4F"),
        ("Cinza ardósia", "#708090"),
        ("Cinza claro", "#D3D3D3"),
        ("Cinza escuro", "#A9A9A9"),
        ("Cinza fosco", "#696969"),
        ("Cinza médio", "#DCDCDC"),
        ("Cinza", "#808080"),
        ("Cobre", "#B87333"),
        ("Concha", "#FFF5EE"),
        ("Coral claro", "#F08080"),
        ("Coral", "#FF7F50"),
        ("Couro", "#F0DC82"),
        ("Creme de marisco", "#FFE4C4"),
        ("Creme de menta", "#F5FFFA"),
        ("Creme", "#FFFDD0"),
        ("Dourado escuro", "#B8860B"),
        ("Dourado pálido", "#EEE8AA"),
        ("Dourado", "#DAA520"),
        ("Ébano", "#555D50"),
        ("Eminência", "#6C3082"),
        ("Escarlate", "#FF2400"),
        ("Esmeralda", "#50C878"),
        ("Eucalipto", "#44D7A8"),
        ("Fandango", "#B53389"),
        ("Feldspato", "#FDD5B1"),
        ("Ferrugem", "#B7410E"),
        ("Flerte", "#A2006D"),
        ("Fúcsia", "#FF00FF"),
        ("Fuligem", "#3D2B1F"),
        ("Glicínia", "#C9A0DC"),
        ("Glitter", "#E6E8FA"),
        ("Grená", "#831D1C"),
        ("Heliotrópio", "#DF73FF"),
        ("Herbal", "#2E8B57"),
        ("Independência", "#4C516D"),
        ("Índigo", "#4B0082"),
        ("Iris", "#5A4FCF"),
        ("Jade", "#00A86B"),
        ("Jambo", "#FF4500"),
        ("Jasmine", "#F8DE7E"),
        ("Kiwi", "#8EE53F"),
        ("Laranja claro", "#FFB84D"),
        ("Laranja escuro", "#FF8C00"),
        ("Laranja", "#FFA500"),
        ("Lavanda avermelhada", "#FFF0F5"),
        ("Lavanda", "#E6E6FA"),
        ("Lilás", "#C8A2C8"),
        ("Lima", "#FDE910"),
        ("Limão", "#00FF00"),
        ("Linho", "#FAF0E6"),
        ("Madeira", "#DEB887"),
        ("Magenta escuro", "#8B008B"),
        ("Magenta", "#FF00FF"),
        ("Malva", "#E0B0FF"),
        ("Mamão batido", "#FFEFD5"),
        ("Maná", "#F0FFF0"),
        ("Marfim", "#FFFFF0"),
        ("Marrom amarelado", "#F4A460"),
        ("Marrom claro", "#A52A2A"),
        ("Marrom rosado", "#BC8F8F"),
        ("Marrom sela", "#8B4513"),
        ("Marrom", "#964B00"),
        ("Milho Claro", "#FFF8DC"),
        ("Milho", "#FBEC5D"),
        ("Mocassim", "#FFE4B5"),
        ("Mostarda", "#FFDB58"),
        ("Naval", "#000080"),
        ("Neve", "#FFFAFA"),
        ("Nyanza", "#E9FFDB"),
        ("Ocre", "#CC7722"),
        ("Oliva escura", "#556B2F"),
        ("Oliva parda", "#6B8E23"),
        ("Oliva", "#808000"),
        ("Orquídea escura", "#9932CC"),
        ("Orquídea média", "#BA55D3"),
        ("Orquídea", "#DA70D6"),
        ("Ouro", "#FFD700"),
        ("Pardo escuro", "#CC6600"),
        ("Pardo", "#CD853F"),
        ("Pêssego", "#FFDAB9"),
        ("Prata", "#C0C0C0"),
        ("Preto", "#000000"),
        ("Púrpura média", "#9370DB"),
        ("Púrpura", "#800080"),
        ("Quantum", "#111111"),
        ("Quartzo", "#51484F"),
        ("Renda antiga", "#FDF5E6"),
        ("Rosa amoroso", "#CD69CD"),
        ("Rosa brilhante", "#FF007F"),
        ("Rosa Choque", "#FC0FC0"),
        ("Rosa claro", "#FFB6C1"),
        ("Rosa danação", "#DA69A1"),
        ("Rosa embaçado", "#FFE4E1"),
        ("Rosa forte", "#FF69B4"),
        ("Rosa profundo", "#FF1493"),
        ("Rosa", "#FFCBDB"),
        ("Roxo brasilis", "#8A008A"),
        ("Roxo", "#993399"),
        ("Rútilo", "#6D351A"),
        ("Salmão claro", "#FFA07A"),
        ("Salmão escuro", "#E9967A"),
        ("Salmão", "#FA7F72"),
        ("Sépia", "#705714"),
        ("Siena", "#FF8247"),
        ("Tangerina", "#F28500"),
        ("Terracota", "#E2725B"),
        ("Tijolo refratário", "#B22222"),
        ("Tomate", "#FF6347"),
        ("Triássico", "#FF2401"),
        ("Trigo", "#F5DEB3"),
        ("Turquesa escura", "#00CED1"),
        ("Turquesa média", "#48D1CC"),
        ("Turquesa pálida", "#AFEEEE"),
        ("Turquesa", "#40E0D0"),
        ("Urucum", "#EC2300"),
        ("Verde amarelado", "#9ACD32"),
        ("Verde claro", "#90EE90"),
        ("Verde escuro", "#006400"),
        ("Verde espectro", "#00FF00"),
        ("Verde floresta", "#228B22"),
        ("Verde fluorescente", "#CCFF33"),
        ("Verde grama", "#7CFC00"),
        ("Verde lima", "#32CD32"),
        ("Verde mar claro", "#20B2AA"),
        ("Verde mar escuro", "#8FBC8F"),
        ("Verde mar médio", "#3CB371"),
        ("Verde militar", "#78866B"),
        ("Verde pálido", "#98FB98"),
        ("Verde Paris", "#7FFF00"),
        ("Verde primavera médio", "#00FA9A"),
        ("Verde primavera", "#00FF7F"),
        ("Verde-azulado", "#008080"),
        ("Verde", "#008000"),
        ("Vermelho enegrecido", "#550000"),
        ("Vermelho escuro", "#8B0000"),
        ("Vermelho indiano", "#CD5C5C"),
        ("Vermelho violeta médio", "#C71585"),
        ("Vermelho violeta pálido", "#DB7093"),
        ("Vermelho violeta", "#D02090"),
        ("Vermelho", "#FF0000"),
        ("Violeta claro", "#F8CBF8"),
        ("Violeta escuro", "#9400D3"),
        ("Violeta", "#EE82EE"),
        ("Zinco", "#E2DDF0"),
    ))

    safe_colors = (
        'preto', 'marrom', 'verde', 'azul escuro', 'verde escuro',
        'roxo', 'laranja', 'verde claro', 'azul', 'rosa', 'violeta',
        'cinza', 'amarelo', 'magenta', 'ciano', 'branco',
    )



import sys
if __name__ == '__main__':
    import pytest
    sys.exit(pytest.main())
else:
    import sys, pytest
    sys.modules['py.test'] = pytest

# for more API entry points see the 'tests' definition
# in __init__.py



from django import VERSION as DJANGO_VERSION
from django.db.backends.postgresql_psycopg2.base import (
    DatabaseCreation,
    DatabaseWrapper
)


class DatabaseWrapper(DatabaseWrapper):
    UNTOUCHED = 1

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.creation = ViewDatabaseCreation(self)
        self.schema = self.UNTOUCHED

    def _cursor(self, name=None):
        from ..manager import get_schema

        cursor = super()._cursor()
        schema = get_schema()
        if schema != self.schema:
            self.schema = schema
            if schema:
                cursor.execute("SET search_path = %s, public", [schema])
            else:
                cursor.execute("SET search_path = public")
        return cursor

    def reset_schema(self):
        self.schema = self.UNTOUCHED


class ViewDatabaseCreation(DatabaseCreation):
    def _sql_for_inline_fk_refs(self, field, known_models, style, model=None):
        # Hack to point references to the correct table.
        view_table = field.rel.to._meta.db_table
        if getattr(field.rel.to._meta, "_is_view", False):
            if field.rel.field_name == field.rel.to._meta.pk.name:
                # Base model might not be created
                if field.rel.to._meta._base_model in known_models:
                    base_table = field.rel.to._meta._base_model._meta.db_table
                    field.rel.to._meta.db_table = base_table
                else:
                    output = []
                    pending = True
                    return output, pending
            else:
                version_table = field.rel.to._meta._version_model._meta.db_table
                field.rel.to._meta.db_table = version_table

        # Create the references
        if DJANGO_VERSION < (1, 6):
            ret_val = super().sql_for_inline_foreign_key_references(field, known_models, style)
        else:
            ret_val = super().sql_for_inline_foreign_key_references(model, field, known_models, style)

        # Restore the db_table value
        field.rel.to._meta.db_table = view_table

        return ret_val

    if DJANGO_VERSION < (1, 6):

        def sql_for_inline_foreign_key_references(self, field, known_models, style):
            return self._sql_for_inline_fk_refs(field, known_models, style)

    else:

        def sql_for_inline_foreign_key_references(
            self, model, field, known_models, style
        ):
            return self._sql_for_inline_fk_refs(field, known_models, style, model)

    def sql_for_pending_references(self, model, style, pending_references):
        from ..fields import FKToVersion

        # Hack to create references after the fact.
        changed = []
        if model in pending_references:
            for rel_class, field in pending_references[model]:
                if getattr(rel_class, "_is_view", False):
                    changed.append((rel_class, rel_class._meta.db_table))

                    # which table do we point to
                    if isinstance(field, FKToVersion):
                        version_table = rel_class._meta._version_model._meta.db_table
                        rel_class._meta.db_table = version_table
                    else:
                        base_table = rel_class._meta._base_model._meta.db_table
                        rel_class._meta.db_table = base_table

        # Create the references
        ret_val = super().sql_for_pending_references(
            model, style, pending_references
        )

        # Restore the db_table values
        for rel_class, table in changed:
            rel_class._meta.db_table = table

        return ret_val



# Generated by Django 2.2.16 on 2021-01-22 23:45

import django.contrib.postgres.fields.jsonb
import django.core.validators
from django.db import migrations, models
import django.db.models.deletion
import django.db.models.manager
import django.utils.timezone
import scarlet.assets.fields
import scarlet.cms.fields
import scarlet.pagebuilder.models
import scarlet.versioning.fields


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('assets', '0004_auto_20201027_0949'),
    ]

    operations = [
        migrations.CreateModel(
            name='Page',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('is_published', models.BooleanField(default=False, editable=False)),
                ('created_date', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('v_last_save', models.DateTimeField(editable=False, null=True)),
                ('state', models.CharField(choices=[('published', 'published'), ('scheduled', 'scheduled'), ('draft', 'draft'), ('archived', 'archived')], editable=False, max_length=50)),
                ('last_scheduled', models.DateTimeField(editable=False, null=True)),
                ('date_published', models.DateTimeField(editable=False, null=True)),
                ('user_published', models.CharField(editable=False, max_length=255, null=True)),
                ('vid', models.PositiveIntegerField(editable=False, unique=True)),
                ('object_id', models.PositiveIntegerField(editable=False)),
                ('title', models.CharField(max_length=255)),
                ('internal_name', models.CharField(blank=True, help_text='For internal use only.', max_length=255, null=True)),
                ('slug', models.SlugField(max_length=255)),
                ('description', models.CharField(blank=True, max_length=500, null=True)),
                ('keywords', models.CharField(blank=True, max_length=255, null=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('og_image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('og_image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
            ],
            options={
                'managed': False,
                'db_table': 'pagebuilder_page_base',
            },
            bases=(models.Model, scarlet.pagebuilder.models.PageReferences),
            managers=[
                ('objects', django.db.models.manager.Manager()),
                ('normal', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='BaseCarouselItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('title', models.CharField(max_length=255)),
                ('subtitle', models.CharField(blank=True, max_length=255, null=True)),
                ('button_url', models.URLField(blank=True, null=True)),
                ('button_text', models.CharField(blank=True, max_length=100, null=True)),
                ('image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('image', scarlet.assets.fields.AssetsFileField(blank=True, denormalize=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseCarouselModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('button_url', models.URLField(blank=True, null=True)),
                ('button_text', models.CharField(blank=True, max_length=100, null=True)),
                ('left_arrow_active', models.BooleanField(default=False)),
                ('right_arrow_active', models.BooleanField(default=False)),
                ('loop_carousel', models.BooleanField(default=False)),
                ('display_arrow_at_end', models.BooleanField(default=False, help_text="If the carousel doesn't loop, then don't display an arrow when a user reaches either end of the carousel.")),
                ('display_slide_indicator', models.BooleanField(default=False)),
                ('visible_image', models.IntegerField(validators=[django.core.validators.MinValueValidator(1), django.core.validators.MaxValueValidator(5)])),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseFAQItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('question', models.CharField(max_length=255)),
                ('answer', scarlet.cms.fields.HTMLTextField()),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseFAQModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('question_text_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255)),
                ('answer_text_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseGalleryImage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('image', scarlet.assets.fields.AssetsFileField(blank=True, denormalize=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseHeaderModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('header', models.CharField(blank=True, max_length=255, null=True, verbose_name='Header Text')),
                ('header_text_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255)),
                ('subheader', models.CharField(blank=True, max_length=255, null=True, verbose_name='Subheader Text')),
                ('subheader_text_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseHeroModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('header', models.CharField(blank=True, max_length=255, null=True, verbose_name='Header Text')),
                ('header_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255, verbose_name='Header Text Alignment')),
                ('subheader', models.CharField(blank=True, max_length=255, null=True, verbose_name='Subheader Text')),
                ('subheader_text_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255, verbose_name='Subheader Text Alignment')),
                ('button_url', models.URLField(blank=True, null=True)),
                ('button_text', models.CharField(blank=True, max_length=100, null=True)),
                ('button_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center'), ('right', 'Right')], default='left', max_length=255, verbose_name='Button Alignment')),
                ('image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('image', scarlet.assets.fields.AssetsFileField(blank=True, denormalize=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseIconListItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('text', scarlet.cms.fields.HTMLTextField(blank=True, null=True)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('icon_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('icon_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('icon', scarlet.assets.fields.AssetsFileField(denormalize=False, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset')),
            ],
        ),
        migrations.CreateModel(
            name='BaseIconListModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('title', models.CharField(max_length=255)),
                ('subtitle', models.CharField(blank=True, max_length=255, null=True)),
                ('footnote', models.CharField(blank=True, max_length=255, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseImageGalleryModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseLocationItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('title', models.CharField(max_length=255)),
                ('address1', models.CharField(help_text='Street address, P.O. Box', max_length=255)),
                ('address2', models.CharField(blank=True, help_text='Apartment, suite, unit, building, floor, etc', max_length=255, null=True)),
                ('city', models.CharField(blank=True, max_length=255, null=True)),
                ('state', models.CharField(blank=True, max_length=255, null=True, verbose_name='State/Province/Region')),
                ('zip_code', models.CharField(blank=True, max_length=10, null=True, verbose_name='Zip/Postal Code')),
                ('country', models.CharField(blank=True, max_length=100, null=True)),
                ('hours', scarlet.cms.fields.HTMLTextField(blank=True, null=True)),
                ('phone_number', models.CharField(blank=True, max_length=20, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseLocationModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='BaseTwoColumnModule',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('last_save', models.DateTimeField(editable=False)),
                ('order', scarlet.cms.fields.OrderField(db_index=True, default=0)),
                ('left_header', models.CharField(blank=True, max_length=255, null=True, verbose_name='Header Text')),
                ('left_header_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center')], default='left', max_length=255, verbose_name='Left Header Text Alignment')),
                ('left_description', scarlet.cms.fields.HTMLTextField(blank=True, null=True)),
                ('left_description_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center')], default='left', max_length=255, verbose_name='Left Description Text Alignment')),
                ('left_button_url', models.URLField(blank=True, null=True)),
                ('left_button_text', models.CharField(blank=True, max_length=100, null=True)),
                ('left_button_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center')], default='left', max_length=255, verbose_name='Left Button Alignment')),
                ('right_header', models.CharField(blank=True, max_length=255, null=True, verbose_name='Header Text')),
                ('right_header_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center')], default='left', max_length=255, verbose_name='Right Header Text Alignment')),
                ('right_description', scarlet.cms.fields.HTMLTextField(blank=True, null=True)),
                ('right_description_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center')], default='left', max_length=255, verbose_name='Right Description Text Alignment')),
                ('right_button_url', models.URLField(blank=True, null=True)),
                ('right_button_text', models.CharField(blank=True, max_length=100, null=True)),
                ('right_button_align', models.CharField(choices=[('left', 'Left'), ('center', 'Center')], default='left', max_length=255, verbose_name='Right Button Alignment')),
                ('left_image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('left_image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('right_image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('right_image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('left_image', scarlet.assets.fields.AssetsFileField(blank=True, denormalize=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset')),
                ('right_image', scarlet.assets.fields.AssetsFileField(blank=True, denormalize=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Page_base',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_published', models.BooleanField(default=False, editable=False)),
                ('created_date', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('v_last_save', models.DateTimeField(editable=False, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Page_version',
            fields=[
                ('last_save', models.DateTimeField(editable=False)),
                ('state', models.CharField(choices=[('published', 'published'), ('scheduled', 'scheduled'), ('draft', 'draft'), ('archived', 'archived')], editable=False, max_length=50)),
                ('last_scheduled', models.DateTimeField(editable=False, null=True)),
                ('date_published', models.DateTimeField(editable=False, null=True)),
                ('user_published', models.CharField(editable=False, max_length=255, null=True)),
                ('title', models.CharField(max_length=255)),
                ('internal_name', models.CharField(blank=True, help_text='For internal use only.', max_length=255, null=True)),
                ('slug', models.SlugField(max_length=255)),
                ('description', models.CharField(blank=True, max_length=500, null=True)),
                ('keywords', models.CharField(blank=True, max_length=255, null=True)),
                ('created', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('modified', models.DateTimeField(auto_now=True)),
                ('vid', models.AutoField(primary_key=True, serialize=False)),
                ('og_image_urls', django.contrib.postgres.fields.jsonb.JSONField(default=dict, editable=False)),
                ('og_image_alt_text', models.CharField(blank=True, max_length=255, null=True)),
                ('object', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='version_version', to='pagebuilder.Page_base')),
                ('og_image', scarlet.assets.fields.AssetsFileField(blank=True, denormalize=False, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='+', to='assets.Asset', verbose_name='Open Graph Image')),
            ],
            options={
                'managed': True,
            },
            bases=(models.Model, scarlet.pagebuilder.models.PageVersionReferences),
            managers=[
                ('objects', django.db.models.manager.Manager()),
                ('normal', django.db.models.manager.Manager()),
            ],
        ),
        migrations.CreateModel(
            name='TwoColumnModule',
            fields=[
                ('basetwocolumnmodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseTwoColumnModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_two_column_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.basetwocolumnmodule',),
        ),
        migrations.CreateModel(
            name='LocationModule',
            fields=[
                ('baselocationmodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseLocationModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_location_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.baselocationmodule',),
        ),
        migrations.CreateModel(
            name='LocationItem',
            fields=[
                ('baselocationitem_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseLocationItem')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='pagebuilder.LocationModule')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.baselocationitem',),
        ),
        migrations.CreateModel(
            name='ImageGalleryModule',
            fields=[
                ('baseimagegallerymodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseImageGalleryModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_gallery_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.baseimagegallerymodule',),
        ),
        migrations.CreateModel(
            name='IconListModule',
            fields=[
                ('baseiconlistmodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseIconListModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_icon_list_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.baseiconlistmodule',),
        ),
        migrations.CreateModel(
            name='IconListItem',
            fields=[
                ('baseiconlistitem_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseIconListItem')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='pagebuilder.IconListModule')),
            ],
            bases=('pagebuilder.baseiconlistitem',),
        ),
        migrations.CreateModel(
            name='HeroModule',
            fields=[
                ('baseheromodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseHeroModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_hero_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.baseheromodule',),
        ),
        migrations.CreateModel(
            name='HeaderModule',
            fields=[
                ('baseheadermodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseHeaderModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_header_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.baseheadermodule',),
        ),
        migrations.CreateModel(
            name='GalleryImage',
            fields=[
                ('basegalleryimage_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseGalleryImage')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='pagebuilder.ImageGalleryModule')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.basegalleryimage',),
        ),
        migrations.CreateModel(
            name='FAQModule',
            fields=[
                ('basefaqmodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseFAQModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_faq_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.basefaqmodule',),
        ),
        migrations.CreateModel(
            name='FAQItem',
            fields=[
                ('basefaqitem_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseFAQItem')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='pagebuilder.FAQModule')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.basefaqitem',),
        ),
        migrations.CreateModel(
            name='CarouselModule',
            fields=[
                ('basecarouselmodule_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseCarouselModule')),
                ('page', scarlet.versioning.fields.FKToVersion(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='page_carousel_modules', to='pagebuilder.Page_version')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.basecarouselmodule',),
        ),
        migrations.CreateModel(
            name='CarouselItem',
            fields=[
                ('basecarouselitem_ptr', models.OneToOneField(auto_created=True, on_delete=django.db.models.deletion.CASCADE, parent_link=True, primary_key=True, serialize=False, to='pagebuilder.BaseCarouselItem')),
                ('module', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='pagebuilder.CarouselModule')),
            ],
            options={
                'abstract': False,
            },
            bases=('pagebuilder.basecarouselitem',),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import os
import pkgutil

_replaces = pkgutil.get_data(__name__, 'data.bin').decode('utf8').split('\x00')

def unidecode(txt):
    chars = []
    for ch in txt:
        codepoint = ord(ch)

        if not codepoint:
            chars.append('\x00')
            continue

        try:
            chars.append(_replaces[codepoint-1])
        except IndexError:
            pass
    return "".join(chars)



from io import StringIO as TextIO
from io import BytesIO as BytesIO
from typing import Any, AnyStr, Callable, Generic, IO, List, Optional, Text, Tuple, TypeVar, Union, overload
from typing_extensions import Final
import sys

_T = TypeVar("_T")

class FDCapture(Generic[AnyStr]):
    def __init__(self, targetfd: int, tmpfile: Optional[IO[AnyStr]] = ..., now: bool = ..., patchsys: bool = ...) -> None: ...
    def start(self) -> None: ...
    def done(self) -> IO[AnyStr]: ...
    def writeorg(self, data: AnyStr) -> None: ...

class StdCaptureFD:
    def __init__(
        self,
        out: Union[bool, IO[str]] = ...,
        err: Union[bool, IO[str]] = ...,
        mixed: bool = ...,
        in_: bool = ...,
        patchsys: bool = ...,
        now: bool = ...,
    ) -> None: ...
    @classmethod
    def call(cls, func: Callable[..., _T], *args: Any, **kwargs: Any) -> Tuple[_T, str, str]: ...
    def reset(self) -> Tuple[str, str]: ...
    def suspend(self) -> Tuple[str, str]: ...
    def startall(self) -> None: ...
    def resume(self) -> None: ...
    def done(self, save: bool = ...) -> Tuple[IO[str], IO[str]]: ...
    def readouterr(self) -> Tuple[str, str]: ...

class StdCapture:
    def __init__(
        self,
        out: Union[bool, IO[str]] = ...,
        err: Union[bool, IO[str]] = ...,
        in_: bool = ...,
        mixed: bool = ...,
        now: bool = ...,
    ) -> None: ...
    @classmethod
    def call(cls, func: Callable[..., _T], *args: Any, **kwargs: Any) -> Tuple[_T, str, str]: ...
    def reset(self) -> Tuple[str, str]: ...
    def suspend(self) -> Tuple[str, str]: ...
    def startall(self) -> None: ...
    def resume(self) -> None: ...
    def done(self, save: bool = ...) -> Tuple[IO[str], IO[str]]: ...
    def readouterr(self) -> Tuple[IO[str], IO[str]]: ...

# XXX: The type here is not exactly right. If f is IO[bytes] and
# encoding is not None, returns some weird hybrid, not exactly IO[bytes].
def dupfile(
    f: IO[AnyStr],
    mode: Optional[str] = ...,
    buffering: int = ...,
    raising: bool = ...,
    encoding: Optional[str] = ...,
) -> IO[AnyStr]: ...
def get_terminal_width() -> int: ...
def ansi_print(
    text: Union[str, Text],
    esc: Union[Union[str, Text], Tuple[Union[str, Text], ...]],
    file: Optional[IO[Any]] = ...,
    newline: bool = ...,
    flush: bool = ...,
) -> None: ...
def saferepr(obj, maxsize: int = ...) -> str: ...

class TerminalWriter:
    stringio: TextIO
    encoding: Final[str]
    hasmarkup: bool
    def __init__(self, file: Optional[IO[str]] = ..., stringio: bool = ..., encoding: Optional[str] = ...) -> None: ...
    @property
    def fullwidth(self) -> int: ...
    @fullwidth.setter
    def fullwidth(self, value: int) -> None: ...
    @property
    def chars_on_current_line(self) -> int: ...
    @property
    def width_of_current_line(self) -> int: ...
    def markup(
        self,
        text: str,
        *,
        black: int = ..., red: int = ..., green: int = ..., yellow: int = ..., blue: int = ..., purple: int = ...,
        cyan: int = ..., white: int = ..., Black: int = ..., Red: int = ..., Green: int = ..., Yellow: int = ...,
        Blue: int = ..., Purple: int = ..., Cyan: int = ..., White: int = ..., bold: int = ..., light: int = ...,
        blink: int = ..., invert: int = ...,
    ) -> str: ...
    def sep(
        self,
        sepchar: str,
        title: Optional[str] = ...,
        fullwidth: Optional[int] = ...,
        *,
        black: int = ..., red: int = ..., green: int = ..., yellow: int = ..., blue: int = ..., purple: int = ...,
        cyan: int = ..., white: int = ..., Black: int = ..., Red: int = ..., Green: int = ..., Yellow: int = ...,
        Blue: int = ..., Purple: int = ..., Cyan: int = ..., White: int = ..., bold: int = ..., light: int = ...,
        blink: int = ..., invert: int = ...,
    ) -> None: ...
    def write(
        self,
        msg: str,
        *,
        black: int = ..., red: int = ..., green: int = ..., yellow: int = ..., blue: int = ..., purple: int = ...,
        cyan: int = ..., white: int = ..., Black: int = ..., Red: int = ..., Green: int = ..., Yellow: int = ...,
        Blue: int = ..., Purple: int = ..., Cyan: int = ..., White: int = ..., bold: int = ..., light: int = ...,
        blink: int = ..., invert: int = ...,
    ) -> None: ...
    def line(
        self,
        s: str = ...,
        *,
        black: int = ..., red: int = ..., green: int = ..., yellow: int = ..., blue: int = ..., purple: int = ...,
        cyan: int = ..., white: int = ..., Black: int = ..., Red: int = ..., Green: int = ..., Yellow: int = ...,
        Blue: int = ..., Purple: int = ..., Cyan: int = ..., White: int = ..., bold: int = ..., light: int = ...,
        blink: int = ..., invert: int = ...,
    ) -> None: ...
    def reline(
        self,
        line: str,
        *,
        black: int = ..., red: int = ..., green: int = ..., yellow: int = ..., blue: int = ..., purple: int = ...,
        cyan: int = ..., white: int = ..., Black: int = ..., Red: int = ..., Green: int = ..., Yellow: int = ...,
        Blue: int = ..., Purple: int = ..., Cyan: int = ..., White: int = ..., bold: int = ..., light: int = ...,
        blink: int = ..., invert: int = ...,
    ) -> None: ...



from scarlet.cms import bundles, views, options
from scarlet.cms.sites import site
from scarlet.cms.forms import search_form

from .models import Page
from .forms import (
    PAGE_EDIT_FORMSETS,
    PAGE_EDIT_FIELDSET,
    MODULE_COMBINED_FORMSET,
    EmptyPageForm,
    PageForm,
)


# class CarouselBundle(bundles.Bundle):
#     dashboard = (("main", "Carousels"),)

#     main = views.ListView()
#     edit = views.FormView(formsets={"Items": ItemInlineFormset()})
#     add = edit

#     class Meta:
#         model = Carousel
#         primary_model_bundle = True


class ContentModuleBundle(bundles.ObjectOnlyBundle):
    navigation = bundles.PARENT

    main = views.FormView(
        form_class=EmptyPageForm,
        formsets=PAGE_EDIT_FORMSETS,
        fieldsets=PAGE_EDIT_FIELDSET,
        combined_formset_defs=MODULE_COMBINED_FORMSET,
    )

    class Meta:
        model = Page


class PageEditBundle(bundles.VersionedObjectOnlyBundle):
    navigation = (
        ("meta", "Meta"),
        ("content_modules", "Content Modules"),
    )

    meta = views.FormView(form_class=PageForm)
    content_modules = ContentModuleBundle.as_subbundle(name="content_modules")
    main = content_modules

    class Meta:
        model = Page


class PageAddBundle(bundles.VersionedObjectOnlyBundle):
    main = views.FormView(form_class=PageForm, redirect_to_view="edit")

    class Meta:
        model = Page


class PageBundle(bundles.VersionedBundle):
    dashboard = (("main", "Page"),)

    edit = PageEditBundle.as_subbundle(name="page_edit")
    add = PageAddBundle.as_subbundle(name="page_add")

    class Meta:
        model = Page
        primary_model_bundle = True


class PagesDashboardBundle(bundles.BlankBundle):
    dashboard = (
        ("page", "Page"),
        ("carousel", "Carousel"),
    )

    page = PageBundle.as_subbundle(name="Page", title="Page", title_plural="Page")


site.register(
    "pages", PagesDashboardBundle(name="pages", title="Page", title_plural="Pages")
)



from threading import local

from django.db import connections, models

_mode = local()


class ContextState:
    """
    Checks the current context every
    time the query is evaluated.
    """

    def prepare(self):
        return self

    def _prepare(self, val=None):
        return self

    def as_sql(self, qn=None, compiler=None, connection=None):
        current_state = getattr(_mode, "current_state", None)
        if not current_state:
            return '"state"', tuple()
        else:
            return "%s", (current_state,)


def activate(state):
    """
    Activate a state in this thread.
    """

    _mode.current_state = state
    _mode.schema = state


def deactivate():
    """
    Deactivate a state in this thread.
    """

    if hasattr(_mode, "current_state"):
        del _mode.current_state
    if hasattr(_mode, "schema"):
        del _mode.schema

    for k in connections:
        con = connections[k]
        if hasattr(con, "reset_schema"):
            con.reset_schema()


def get_schema():
    return getattr(_mode, "schema", None)


class VersionManager(models.Manager):
    """
    Default Manager for version models.

    Looks up the current state and filters on it.
    """

    use_for_related_fields = True

    def get_query_set(self):
        return self.get_queryset()

    def get_queryset(self):
        try:
            qs = super().get_query_set()
        except AttributeError:
            qs = super().get_queryset()
        qs = qs.filter(state=ContextState())
        return qs


class SwitchSchema:
    """
    Context manager for switching schema.
    """

    def __init__(self, schema):
        self.old_schema = get_schema()
        self.schema = schema

    def __enter__(self):
        _mode.schema = self.schema

    def __exit__(self, etype, value, traceback):
        _mode.schema = self.old_schema


class SwitchSchemaManager(SwitchSchema):
    def __enter__(self):
        _mode.schema = self.schema
        _mode.current_state = self.schema

    def __exit__(self, etype, value, traceback):
        _mode.schema = self.old_schema
        _mode.current_state = self.old_schema



# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/django_coverage_plugin/blob/master/NOTICE.txt

"""Django Template Coverage Plugin"""

from .plugin import DjangoTemplatePlugin
from .plugin import DjangoTemplatePluginException       # noqa


def coverage_init(reg, options):
    reg.add_file_tracer(DjangoTemplatePlugin())



from django.test import TestCase

from scarlet.versioning import manager

from . import models


class ManagerTests(TestCase):
    fixtures = ("test_data.json",)

    def testActivate(self):
        manager.deactivate()

        book = models.Book.objects.get(vid=1)
        book.publish()
        self.assertTrue(
            models.Book.objects.filter(
                state=models.Book.PUBLISHED, object_id=book.object_id
            )
        )

        manager.activate("draft")
        self.assertFalse(
            models.Book.objects.filter(
                state=models.Book.PUBLISHED, object_id=book.object_id
            )
        )
        schema = manager.get_schema()
        self.assertEqual(schema, "draft")

        manager.deactivate()
        self.assertTrue(
            models.Book.objects.filter(
                state=models.Book.PUBLISHED, object_id=book.object_id
            )
        )
        schema = manager.get_schema()
        self.assertEqual(schema, None)

    def testLazyEval(self):
        manager.deactivate()

        book = models.Book.objects.get(vid=1)
        book.publish()

        manager.activate("published")
        qs = models.Book.objects.filter(pk=1)
        self.assertEqual(qs.count(), 1)
        manager.deactivate()

        manager.activate("draft")
        self.assertEqual(qs.count(), 1)
        manager.deactivate()

        self.assertEqual(qs.count(), 2)



from .. import Provider as AutomotiveProvider


class Provider(AutomotiveProvider):

    # From: https://pt.wikipedia.org/wiki/Matr%C3%ADculas_de_autom%C3%B3veis_em_Portugal
    license_formats = (
        '##-##-??',
        '##-??-##',
        '??-##-##',
    )



from functools import wraps

from django.utils.decorators import available_attrs


class CacheMethodResponse:
    """
    Decorator that can be used to cache a method of a
    class that implements CacheMixin
    """

    def __call__(self, func):
        this = self

        @wraps(func, assigned=available_attrs(func))
        def inner(self, request, *args, **kwargs):
            response = None
            self.cache_middleware = None
            if self.should_cache():
                prefix = f"{self.get_cache_version()}:{self.get_cache_prefix()}"

                # Using middleware here since that is what the decorator uses
                # internally and it avoids making this code all complicated with
                # all sorts of wrappers.
                self.set_cache_middleware(self.cache_time, prefix)
                response = self.cache_middleware.process_request(self.request)
            else:
                self.set_do_not_cache()

            if not response:
                response = func(self, request, *args, **kwargs)

            return self._finalize_cached_response(request, response)

        return inner


cache_method_response = CacheMethodResponse



#
# Copyright (C) 2009-2020 the sqlparse authors and contributors
# <see AUTHORS file>
#
# This module is part of python-sqlparse and is released under
# the BSD License: https://opensource.org/licenses/BSD-3-Clause

import re

from sqlparse import sql, tokens as T
from sqlparse.utils import split_unquoted_newlines


class StripCommentsFilter:

    @staticmethod
    def _process(tlist):
        def get_next_comment():
            # TODO(andi) Comment types should be unified, see related issue38
            return tlist.token_next_by(i=sql.Comment, t=T.Comment)

        def _get_insert_token(token):
            """Returns either a whitespace or the line breaks from token."""
            # See issue484 why line breaks should be preserved.
            m = re.search(r'((\r\n|\r|\n)+) *$', token.value)
            if m is not None:
                return sql.Token(T.Whitespace.Newline, m.groups()[0])
            else:
                return sql.Token(T.Whitespace, ' ')

        tidx, token = get_next_comment()
        while token:
            pidx, prev_ = tlist.token_prev(tidx, skip_ws=False)
            nidx, next_ = tlist.token_next(tidx, skip_ws=False)
            # Replace by whitespace if prev and next exist and if they're not
            # whitespaces. This doesn't apply if prev or next is a parenthesis.
            if (prev_ is None or next_ is None
                    or prev_.is_whitespace or prev_.match(T.Punctuation, '(')
                    or next_.is_whitespace or next_.match(T.Punctuation, ')')):
                # Insert a whitespace to ensure the following SQL produces
                # a valid SQL (see #425).
                if prev_ is not None and not prev_.match(T.Punctuation, '('):
                    tlist.tokens.insert(tidx, _get_insert_token(token))
                tlist.tokens.remove(token)
            else:
                tlist.tokens[tidx] = _get_insert_token(token)

            tidx, token = get_next_comment()

    def process(self, stmt):
        [self.process(sgroup) for sgroup in stmt.get_sublists()]
        StripCommentsFilter._process(stmt)
        return stmt


class StripWhitespaceFilter:
    def _stripws(self, tlist):
        func_name = '_stripws_{cls}'.format(cls=type(tlist).__name__)
        func = getattr(self, func_name.lower(), self._stripws_default)
        func(tlist)

    @staticmethod
    def _stripws_default(tlist):
        last_was_ws = False
        is_first_char = True
        for token in tlist.tokens:
            if token.is_whitespace:
                token.value = '' if last_was_ws or is_first_char else ' '
            last_was_ws = token.is_whitespace
            is_first_char = False

    def _stripws_identifierlist(self, tlist):
        # Removes newlines before commas, see issue140
        last_nl = None
        for token in list(tlist.tokens):
            if last_nl and token.ttype is T.Punctuation and token.value == ',':
                tlist.tokens.remove(last_nl)
            last_nl = token if token.is_whitespace else None

            # next_ = tlist.token_next(token, skip_ws=False)
            # if (next_ and not next_.is_whitespace and
            #             token.ttype is T.Punctuation and token.value == ','):
            #     tlist.insert_after(token, sql.Token(T.Whitespace, ' '))
        return self._stripws_default(tlist)

    def _stripws_parenthesis(self, tlist):
        while tlist.tokens[1].is_whitespace:
            tlist.tokens.pop(1)
        while tlist.tokens[-2].is_whitespace:
            tlist.tokens.pop(-2)
        self._stripws_default(tlist)

    def process(self, stmt, depth=0):
        [self.process(sgroup, depth + 1) for sgroup in stmt.get_sublists()]
        self._stripws(stmt)
        if depth == 0 and stmt.tokens and stmt.tokens[-1].is_whitespace:
            stmt.tokens.pop(-1)
        return stmt


class SpacesAroundOperatorsFilter:
    @staticmethod
    def _process(tlist):

        ttypes = (T.Operator, T.Comparison)
        tidx, token = tlist.token_next_by(t=ttypes)
        while token:
            nidx, next_ = tlist.token_next(tidx, skip_ws=False)
            if next_ and next_.ttype != T.Whitespace:
                tlist.insert_after(tidx, sql.Token(T.Whitespace, ' '))

            pidx, prev_ = tlist.token_prev(tidx, skip_ws=False)
            if prev_ and prev_.ttype != T.Whitespace:
                tlist.insert_before(tidx, sql.Token(T.Whitespace, ' '))
                tidx += 1  # has to shift since token inserted before it

            # assert tlist.token_index(token) == tidx
            tidx, token = tlist.token_next_by(t=ttypes, idx=tidx)

    def process(self, stmt):
        [self.process(sgroup) for sgroup in stmt.get_sublists()]
        SpacesAroundOperatorsFilter._process(stmt)
        return stmt


# ---------------------------
# postprocess

class SerializerUnicode:
    @staticmethod
    def process(stmt):
        lines = split_unquoted_newlines(stmt)
        return '\n'.join(line.rstrip() for line in lines)



# Generated by Django 2.2.16 on 2021-01-28 00:04

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pagebuilder', '0002_auto_20210125_1906'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='basecarouselmodule',
            name='display_arrow_at_end',
        ),
        migrations.RemoveField(
            model_name='basecarouselmodule',
            name='display_slide_indicator',
        ),
        migrations.RemoveField(
            model_name='basecarouselmodule',
            name='left_arrow_active',
        ),
        migrations.RemoveField(
            model_name='basecarouselmodule',
            name='loop_carousel',
        ),
        migrations.RemoveField(
            model_name='basecarouselmodule',
            name='right_arrow_active',
        ),
        migrations.RemoveField(
            model_name='basecarouselmodule',
            name='visible_image',
        ),
        migrations.RemoveField(
            model_name='basefaqmodule',
            name='answer_text_align',
        ),
        migrations.RemoveField(
            model_name='basefaqmodule',
            name='question_text_align',
        ),
        migrations.RemoveField(
            model_name='baseheadermodule',
            name='header_text_align',
        ),
        migrations.RemoveField(
            model_name='baseheadermodule',
            name='subheader_text_align',
        ),
        migrations.RemoveField(
            model_name='baseheromodule',
            name='button_align',
        ),
        migrations.RemoveField(
            model_name='baseheromodule',
            name='header_align',
        ),
        migrations.RemoveField(
            model_name='baseheromodule',
            name='subheader_text_align',
        ),
        migrations.RemoveField(
            model_name='basetwocolumnmodule',
            name='left_button_align',
        ),
        migrations.RemoveField(
            model_name='basetwocolumnmodule',
            name='left_description_align',
        ),
        migrations.RemoveField(
            model_name='basetwocolumnmodule',
            name='left_header_align',
        ),
        migrations.RemoveField(
            model_name='basetwocolumnmodule',
            name='right_button_align',
        ),
        migrations.RemoveField(
            model_name='basetwocolumnmodule',
            name='right_description_align',
        ),
        migrations.RemoveField(
            model_name='basetwocolumnmodule',
            name='right_header_align',
        ),
        migrations.AddField(
            model_name='basecarouselitem',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basecarouselitem',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basecarouselmodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basecarouselmodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basefaqitem',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basefaqitem',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basefaqmodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basefaqmodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basegalleryimage',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basegalleryimage',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseheadermodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseheadermodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseheromodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseheromodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseiconlistmodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseiconlistmodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseimagegallerymodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseimagegallerymodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baselocationitem',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baselocationitem',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baselocationmodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baselocationmodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basetwocolumnmodule',
            name='behavior',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basetwocolumnmodule',
            name='style',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



# Generated by Django 2.2.16 on 2021-01-25 19:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('pagebuilder', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='basecarouselitem',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basecarouselmodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basefaqitem',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basefaqmodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basegalleryimage',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseheadermodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseheromodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseiconlistmodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baseimagegallerymodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baselocationitem',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='baselocationmodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basetwocolumnmodule',
            name='internal_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



from django.template import Library, Node, TemplateSyntaxError
from django.template.defaulttags import kwarg_re
from django.utils.encoding import smart_str
from django.utils.safestring import mark_safe

try:
    from django.contrib.auth import get_user_model
except ImportError:
    from django.contrib.auth.models import User
else:
    User = get_user_model()

register = Library()


class URLNode(Node):
    def __init__(self, bundle, view_name, kwargs, asvar):
        self.bundle = bundle
        self.view_name = view_name
        self.kwargs = kwargs
        self.asvar = asvar

    def render(self, context):
        from django.urls.exceptions import NoReverseMatch

        url = ""

        # Resolve vars
        kwargs = dict(
            [
                (smart_str(k, "ascii"), v.resolve(context))
                for k, v in list(self.kwargs.items())
            ]
        )
        follow_parent = kwargs.pop("follow_parent", True)
        bundle = self.bundle.resolve(context)
        url_params = context["url_params"]

        view_name = self.view_name.resolve(context)
        # Try to get the url
        try:
            url = bundle.get_view_url(
                view_name,
                context["user"],
                url_kwargs=kwargs,
                context_kwargs=url_params,
                follow_parent=follow_parent,
            )
        except NoReverseMatch:
            pass

        if self.asvar:
            context[self.asvar] = url
            return ""
        else:
            return url


class ViewNode(Node):
    def __init__(self, bundle, view_name, asvar):
        self.bundle = bundle
        self.view_name = view_name
        self.asvar = asvar

    def render(self, context):

        bundle = self.bundle.resolve(context)
        url_params = context["url_params"]
        view_name = self.view_name.resolve(context)

        # Get the string
        resp = bundle.get_string_from_view(context["request"], view_name, url_params)

        if self.asvar:
            context[self.asvar] = resp
            return ""
        else:
            return resp


class StringNode(Node):
    def __init__(self, viewname, kwargs):
        self.viewname = viewname
        self.kwargs = kwargs

    def render(self, context):

        modules = self.viewname.split(".")
        if len(modules) == 1:
            module = modules[0]
            view = __import__(module, globals(), locals(), [], -1)
        else:
            module = ".".join(modules[:-1])
            view_name = str(modules[-1])
            module = __import__(module, globals(), locals(), [view_name], -1)
            view = getattr(module, view_name)

        kwargs = dict(
            [
                (smart_str(k, "ascii"), v.resolve(context))
                for k, v in list(self.kwargs.items())
            ]
        )

        class_kwargs = {}
        for k, v in list(kwargs.items()):
            if hasattr(view, k):
                class_kwargs[k] = v

        return view.as_string(**class_kwargs)(context["request"], **kwargs)


@register.tag
def render_view(parser, token):
    """
    Return an string version of a View with as_string method.
    First argument is the name of the view. Any other arguments
    should be keyword arguments and will be passed to the view.

    Example:

    {% render_view viewname var1=xx var2=yy %}
    """
    bits = token.split_contents()

    n = len(bits)
    if n < 2:
        raise TemplateSyntaxError("'%s' takes at least one view as argument")

    viewname = bits[1]

    kwargs = {}
    if n > 2:
        for bit in bits[2:]:
            match = kwarg_re.match(bit)
            if not match:
                raise TemplateSyntaxError("Malformed arguments to render_view tag")
            name, value = match.groups()
            if name:
                kwargs[name] = parser.compile_filter(value)

    return StringNode(viewname, kwargs)


@register.tag
def bundle_view(parser, token):
    """
    Returns an string version of a bundle view. This is done by
    calling the `get_string_from_view` method of the provided bundle.

    This tag expects that the request object as well as the
    the original url_params are available in the context.

    Requires two arguments bundle and the name of the view
    you want to render. In addition, this tag also accepts
    the 'as xxx' syntax.

    Example:

    {% bundle_url bundle main_list as html %}
    """

    bits = token.split_contents()
    if len(bits) < 3:
        raise TemplateSyntaxError(
            f"'{bits[0]}' takes at least two arguments: bundle and view_name"
        )

    bundle = parser.compile_filter(bits[1])
    viewname = parser.compile_filter(bits[2])

    asvar = None
    bits = bits[2:]
    if len(bits) >= 2 and bits[-2] == "as":
        asvar = bits[-1]
        bits = bits[:-2]

    return ViewNode(bundle, viewname, asvar)


@register.tag
def bundle_url(parser, token):
    """
    Returns an a url for given a bundle and a view name.
    This is done by calling the `get_view_url` method
    of the provided bundle.

    This tag expects that the request object as well as the
    the original url_params are available in the context.

    Requires two arguments bundle and the name of the view
    you want to render. In addition, this tag also accepts
    the 'as xxx' syntax.

    By default this tag will follow references to
    parent bundles. To stop this from happening pass
    `follow_parent=False`. Any other keyword arguments
    will be used as url keyword arguments.

    If no match is found a blank string will be returned.

    Example:

    {% bundle_url bundle "edit" obj=obj as html %}
    """

    bits = token.split_contents()
    if len(bits) < 3:
        raise TemplateSyntaxError(
            f"'{bits[0]}' takes at least two arguments: bundle and view_name"
        )

    bundle = parser.compile_filter(bits[1])
    viewname = parser.compile_filter(bits[2])

    kwargs = {}
    asvar = None
    bits = bits[2:]
    if len(bits) >= 2 and bits[-2] == "as":
        asvar = bits[-1]
        bits = bits[:-2]

    if len(bits):
        for bit in bits:
            match = kwarg_re.match(bit)
            if not match:
                raise TemplateSyntaxError("Malformed arguments to url tag")
            name, value = match.groups()
            if name:
                kwargs[name] = parser.compile_filter(value)

    return URLNode(bundle, viewname, kwargs, asvar)


@register.simple_tag
def user_url(user, bundle):
    """
    Filter for a user object. Checks if a user has
    permission to change other users.
    """
    if not user:
        return False

    bundle = bundle.admin_site.get_bundle_for_model(User)
    edit = None

    if bundle:
        edit = bundle.get_view_url("main", user)
    return edit


# -- Filters:


@register.filter
def set_label_hidden_formset(field):
    """
    This filter returns `label` HTML code for `AssetsFileFormField` inside formsets.
    For some reasons those fields are hidden when template tries to access to them

    See: `templates/cms/templates/formset.html`
    """
    from scarlet.assets.fields import AssetsFileFormField
    if field.field.__class__ == AssetsFileFormField:
        css_class = ''
        if field.field.required:
            css_class = 'required'
        return mark_safe('<p><label for="{0}" class="{1}">{2}:</label>'.format(field.auto_id, css_class, field.field.label))
    return ''



import functools
import inspect
import os
import sys
import warnings
from collections import defaultdict
from collections import deque
from types import TracebackType
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Generator
from typing import Generic
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import TypeVar
from typing import Union

import attr
import py

import _pytest
from _pytest._code import getfslineno
from _pytest._code.code import FormattedExcinfo
from _pytest._code.code import TerminalRepr
from _pytest._io import TerminalWriter
from _pytest.compat import _format_args
from _pytest.compat import _PytestWrapper
from _pytest.compat import final
from _pytest.compat import get_real_func
from _pytest.compat import get_real_method
from _pytest.compat import getfuncargnames
from _pytest.compat import getimfunc
from _pytest.compat import getlocation
from _pytest.compat import is_generator
from _pytest.compat import NOTSET
from _pytest.compat import order_preserving_dict
from _pytest.compat import overload
from _pytest.compat import safe_getattr
from _pytest.compat import TYPE_CHECKING
from _pytest.config import _PluggyPlugin
from _pytest.config import Config
from _pytest.config.argparsing import Parser
from _pytest.deprecated import FILLFUNCARGS
from _pytest.mark import Mark
from _pytest.mark import ParameterSet
from _pytest.outcomes import fail
from _pytest.outcomes import TEST_OUTCOME
from _pytest.pathlib import absolutepath

if TYPE_CHECKING:
    from typing import Deque
    from typing import NoReturn
    from typing import Type
    from typing_extensions import Literal

    from _pytest import nodes
    from _pytest.main import Session
    from _pytest.python import CallSpec2
    from _pytest.python import Function
    from _pytest.python import Metafunc

    _Scope = Literal["session", "package", "module", "class", "function"]


# The value of the fixture -- return/yield of the fixture function (type variable).
_FixtureValue = TypeVar("_FixtureValue")
# The type of the fixture function (type variable).
_FixtureFunction = TypeVar("_FixtureFunction", bound=Callable[..., object])
# The type of a fixture function (type alias generic in fixture value).
_FixtureFunc = Union[
    Callable[..., _FixtureValue], Callable[..., Generator[_FixtureValue, None, None]]
]
# The type of FixtureDef.cached_result (type alias generic in fixture value).
_FixtureCachedResult = Union[
    Tuple[
        # The result.
        _FixtureValue,
        # Cache key.
        object,
        None,
    ],
    Tuple[
        None,
        # Cache key.
        object,
        # Exc info if raised.
        Tuple["Type[BaseException]", BaseException, TracebackType],
    ],
]


@attr.s(frozen=True)
class PseudoFixtureDef(Generic[_FixtureValue]):
    cached_result = attr.ib(type="_FixtureCachedResult[_FixtureValue]")
    scope = attr.ib(type="_Scope")


def pytest_sessionstart(session: "Session") -> None:
    import _pytest.python
    import _pytest.nodes

    scopename2class.update(
        {
            "package": _pytest.python.Package,
            "class": _pytest.python.Class,
            "module": _pytest.python.Module,
            "function": _pytest.nodes.Item,
            "session": _pytest.main.Session,
        }
    )
    session._fixturemanager = FixtureManager(session)


scopename2class = {}  # type: Dict[str, Type[nodes.Node]]


def get_scope_package(node, fixturedef: "FixtureDef[object]"):
    import pytest

    cls = pytest.Package
    current = node
    fixture_package_name = "{}/{}".format(fixturedef.baseid, "__init__.py")
    while current and (
        type(current) is not cls or fixture_package_name != current.nodeid
    ):
        current = current.parent
    if current is None:
        return node.session
    return current


def get_scope_node(node, scope):
    cls = scopename2class.get(scope)
    if cls is None:
        raise ValueError("unknown scope")
    return node.getparent(cls)


def add_funcarg_pseudo_fixture_def(
    collector, metafunc: "Metafunc", fixturemanager: "FixtureManager"
) -> None:
    # This function will transform all collected calls to functions
    # if they use direct funcargs (i.e. direct parametrization)
    # because we want later test execution to be able to rely on
    # an existing FixtureDef structure for all arguments.
    # XXX we can probably avoid this algorithm  if we modify CallSpec2
    # to directly care for creating the fixturedefs within its methods.
    if not metafunc._calls[0].funcargs:
        # This function call does not have direct parametrization.
        return
    # Collect funcargs of all callspecs into a list of values.
    arg2params = {}  # type: Dict[str, List[object]]
    arg2scope = {}  # type: Dict[str, _Scope]
    for callspec in metafunc._calls:
        for argname, argvalue in callspec.funcargs.items():
            assert argname not in callspec.params
            callspec.params[argname] = argvalue
            arg2params_list = arg2params.setdefault(argname, [])
            callspec.indices[argname] = len(arg2params_list)
            arg2params_list.append(argvalue)
            if argname not in arg2scope:
                scopenum = callspec._arg2scopenum.get(argname, scopenum_function)
                arg2scope[argname] = scopes[scopenum]
        callspec.funcargs.clear()

    # Register artificial FixtureDef's so that later at test execution
    # time we can rely on a proper FixtureDef to exist for fixture setup.
    arg2fixturedefs = metafunc._arg2fixturedefs
    for argname, valuelist in arg2params.items():
        # If we have a scope that is higher than function, we need
        # to make sure we only ever create an according fixturedef on
        # a per-scope basis. We thus store and cache the fixturedef on the
        # node related to the scope.
        scope = arg2scope[argname]
        node = None
        if scope != "function":
            node = get_scope_node(collector, scope)
            if node is None:
                assert scope == "class" and isinstance(collector, _pytest.python.Module)
                # Use module-level collector for class-scope (for now).
                node = collector
        if node and argname in node._name2pseudofixturedef:
            arg2fixturedefs[argname] = [node._name2pseudofixturedef[argname]]
        else:
            fixturedef = FixtureDef(
                fixturemanager=fixturemanager,
                baseid="",
                argname=argname,
                func=get_direct_param_fixture_func,
                scope=arg2scope[argname],
                params=valuelist,
                unittest=False,
                ids=None,
            )
            arg2fixturedefs[argname] = [fixturedef]
            if node is not None:
                node._name2pseudofixturedef[argname] = fixturedef


def getfixturemarker(obj: object) -> Optional["FixtureFunctionMarker"]:
    """Return fixturemarker or None if it doesn't exist or raised
    exceptions."""
    try:
        fixturemarker = getattr(
            obj, "_pytestfixturefunction", None
        )  # type: Optional[FixtureFunctionMarker]
    except TEST_OUTCOME:
        # some objects raise errors like request (from flask import request)
        # we don't expect them to be fixture functions
        return None
    return fixturemarker


# Parametrized fixture key, helper alias for code below.
_Key = Tuple[object, ...]


def get_parametrized_fixture_keys(item: "nodes.Item", scopenum: int) -> Iterator[_Key]:
    """Return list of keys for all parametrized arguments which match
    the specified scope. """
    assert scopenum < scopenum_function  # function
    try:
        callspec = item.callspec  # type: ignore[attr-defined]
    except AttributeError:
        pass
    else:
        cs = callspec  # type: CallSpec2
        # cs.indices.items() is random order of argnames.  Need to
        # sort this so that different calls to
        # get_parametrized_fixture_keys will be deterministic.
        for argname, param_index in sorted(cs.indices.items()):
            if cs._arg2scopenum[argname] != scopenum:
                continue
            if scopenum == 0:  # session
                key = (argname, param_index)  # type: _Key
            elif scopenum == 1:  # package
                key = (argname, param_index, item.fspath.dirpath())
            elif scopenum == 2:  # module
                key = (argname, param_index, item.fspath)
            elif scopenum == 3:  # class
                item_cls = item.cls  # type: ignore[attr-defined]
                key = (argname, param_index, item.fspath, item_cls)
            yield key


# Algorithm for sorting on a per-parametrized resource setup basis.
# It is called for scopenum==0 (session) first and performs sorting
# down to the lower scopes such as to minimize number of "high scope"
# setups and teardowns.


def reorder_items(items: "Sequence[nodes.Item]") -> "List[nodes.Item]":
    argkeys_cache = {}  # type: Dict[int, Dict[nodes.Item, Dict[_Key, None]]]
    items_by_argkey = {}  # type: Dict[int, Dict[_Key, Deque[nodes.Item]]]
    for scopenum in range(0, scopenum_function):
        d = {}  # type: Dict[nodes.Item, Dict[_Key, None]]
        argkeys_cache[scopenum] = d
        item_d = defaultdict(deque)  # type: Dict[_Key, Deque[nodes.Item]]
        items_by_argkey[scopenum] = item_d
        for item in items:
            # cast is a workaround for https://github.com/python/typeshed/issues/3800.
            keys = cast(
                "Dict[_Key, None]",
                order_preserving_dict.fromkeys(
                    get_parametrized_fixture_keys(item, scopenum), None
                ),
            )
            if keys:
                d[item] = keys
                for key in keys:
                    item_d[key].append(item)
    # cast is a workaround for https://github.com/python/typeshed/issues/3800.
    items_dict = cast(
        "Dict[nodes.Item, None]", order_preserving_dict.fromkeys(items, None)
    )
    return list(reorder_items_atscope(items_dict, argkeys_cache, items_by_argkey, 0))


def fix_cache_order(
    item: "nodes.Item",
    argkeys_cache: "Dict[int, Dict[nodes.Item, Dict[_Key, None]]]",
    items_by_argkey: "Dict[int, Dict[_Key, Deque[nodes.Item]]]",
) -> None:
    for scopenum in range(0, scopenum_function):
        for key in argkeys_cache[scopenum].get(item, []):
            items_by_argkey[scopenum][key].appendleft(item)


def reorder_items_atscope(
    items: "Dict[nodes.Item, None]",
    argkeys_cache: "Dict[int, Dict[nodes.Item, Dict[_Key, None]]]",
    items_by_argkey: "Dict[int, Dict[_Key, Deque[nodes.Item]]]",
    scopenum: int,
) -> "Dict[nodes.Item, None]":
    if scopenum >= scopenum_function or len(items) < 3:
        return items
    ignore = set()  # type: Set[Optional[_Key]]
    items_deque = deque(items)
    items_done = order_preserving_dict()  # type: Dict[nodes.Item, None]
    scoped_items_by_argkey = items_by_argkey[scopenum]
    scoped_argkeys_cache = argkeys_cache[scopenum]
    while items_deque:
        no_argkey_group = order_preserving_dict()  # type: Dict[nodes.Item, None]
        slicing_argkey = None
        while items_deque:
            item = items_deque.popleft()
            if item in items_done or item in no_argkey_group:
                continue
            argkeys = order_preserving_dict.fromkeys(
                (k for k in scoped_argkeys_cache.get(item, []) if k not in ignore), None
            )
            if not argkeys:
                no_argkey_group[item] = None
            else:
                slicing_argkey, _ = argkeys.popitem()
                # We don't have to remove relevant items from later in the
                # deque because they'll just be ignored.
                matching_items = [
                    i for i in scoped_items_by_argkey[slicing_argkey] if i in items
                ]
                for i in reversed(matching_items):
                    fix_cache_order(i, argkeys_cache, items_by_argkey)
                    items_deque.appendleft(i)
                break
        if no_argkey_group:
            no_argkey_group = reorder_items_atscope(
                no_argkey_group, argkeys_cache, items_by_argkey, scopenum + 1
            )
            for item in no_argkey_group:
                items_done[item] = None
        ignore.add(slicing_argkey)
    return items_done


def _fillfuncargs(function: "Function") -> None:
    """Fill missing fixtures for a test function, old public API (deprecated)."""
    warnings.warn(FILLFUNCARGS.format(name="pytest._fillfuncargs()"), stacklevel=2)
    _fill_fixtures_impl(function)


def fillfixtures(function: "Function") -> None:
    """Fill missing fixtures for a test function (deprecated)."""
    warnings.warn(
        FILLFUNCARGS.format(name="_pytest.fixtures.fillfixtures()"), stacklevel=2
    )
    _fill_fixtures_impl(function)


def _fill_fixtures_impl(function: "Function") -> None:
    """Internal implementation to fill fixtures on the given function object."""
    try:
        request = function._request
    except AttributeError:
        # XXX this special code path is only expected to execute
        # with the oejskit plugin.  It uses classes with funcargs
        # and we thus have to work a bit to allow this.
        fm = function.session._fixturemanager
        assert function.parent is not None
        fi = fm.getfixtureinfo(function.parent, function.obj, None)
        function._fixtureinfo = fi
        request = function._request = FixtureRequest(function)
        request._fillfixtures()
        # Prune out funcargs for jstests.
        newfuncargs = {}
        for name in fi.argnames:
            newfuncargs[name] = function.funcargs[name]
        function.funcargs = newfuncargs
    else:
        request._fillfixtures()


def get_direct_param_fixture_func(request):
    return request.param


@attr.s(slots=True)
class FuncFixtureInfo:
    # Original function argument names.
    argnames = attr.ib(type=Tuple[str, ...])
    # Argnames that function immediately requires. These include argnames +
    # fixture names specified via usefixtures and via autouse=True in fixture
    # definitions.
    initialnames = attr.ib(type=Tuple[str, ...])
    names_closure = attr.ib(type=List[str])
    name2fixturedefs = attr.ib(type=Dict[str, Sequence["FixtureDef[Any]"]])

    def prune_dependency_tree(self) -> None:
        """Recompute names_closure from initialnames and name2fixturedefs.

        Can only reduce names_closure, which means that the new closure will
        always be a subset of the old one. The order is preserved.

        This method is needed because direct parametrization may shadow some
        of the fixtures that were included in the originally built dependency
        tree. In this way the dependency tree can get pruned, and the closure
        of argnames may get reduced.
        """
        closure = set()  # type: Set[str]
        working_set = set(self.initialnames)
        while working_set:
            argname = working_set.pop()
            # Argname may be smth not included in the original names_closure,
            # in which case we ignore it. This currently happens with pseudo
            # FixtureDefs which wrap 'get_direct_param_fixture_func(request)'.
            # So they introduce the new dependency 'request' which might have
            # been missing in the original tree (closure).
            if argname not in closure and argname in self.names_closure:
                closure.add(argname)
                if argname in self.name2fixturedefs:
                    working_set.update(self.name2fixturedefs[argname][-1].argnames)

        self.names_closure[:] = sorted(closure, key=self.names_closure.index)


class FixtureRequest:
    """A request for a fixture from a test or fixture function.

    A request object gives access to the requesting test context and has
    an optional ``param`` attribute in case the fixture is parametrized
    indirectly.
    """

    def __init__(self, pyfuncitem) -> None:
        self._pyfuncitem = pyfuncitem
        #: Fixture for which this request is being performed.
        self.fixturename = None  # type: Optional[str]
        #: Scope string, one of "function", "class", "module", "session".
        self.scope = "function"  # type: _Scope
        self._fixture_defs = {}  # type: Dict[str, FixtureDef[Any]]
        fixtureinfo = pyfuncitem._fixtureinfo  # type: FuncFixtureInfo
        self._arg2fixturedefs = fixtureinfo.name2fixturedefs.copy()
        self._arg2index = {}  # type: Dict[str, int]
        self._fixturemanager = (
            pyfuncitem.session._fixturemanager
        )  # type: FixtureManager

    @property
    def fixturenames(self) -> List[str]:
        """Names of all active fixtures in this request."""
        result = list(self._pyfuncitem._fixtureinfo.names_closure)
        result.extend(set(self._fixture_defs).difference(result))
        return result

    @property
    def node(self):
        """Underlying collection node (depends on current request scope)."""
        return self._getscopeitem(self.scope)

    def _getnextfixturedef(self, argname: str) -> "FixtureDef[Any]":
        fixturedefs = self._arg2fixturedefs.get(argname, None)
        if fixturedefs is None:
            # We arrive here because of a dynamic call to
            # getfixturevalue(argname) usage which was naturally
            # not known at parsing/collection time.
            assert self._pyfuncitem.parent is not None
            parentid = self._pyfuncitem.parent.nodeid
            fixturedefs = self._fixturemanager.getfixturedefs(argname, parentid)
            # TODO: Fix this type ignore. Either add assert or adjust types.
            #       Can this be None here?
            self._arg2fixturedefs[argname] = fixturedefs  # type: ignore[assignment]
        # fixturedefs list is immutable so we maintain a decreasing index.
        index = self._arg2index.get(argname, 0) - 1
        if fixturedefs is None or (-index > len(fixturedefs)):
            raise FixtureLookupError(argname, self)
        self._arg2index[argname] = index
        return fixturedefs[index]

    @property
    def config(self) -> Config:
        """The pytest config object associated with this request."""
        return self._pyfuncitem.config  # type: ignore[no-any-return] # noqa: F723

    @property
    def function(self):
        """Test function object if the request has a per-function scope."""
        if self.scope != "function":
            raise AttributeError(
                "function not available in {}-scoped context".format(self.scope)
            )
        return self._pyfuncitem.obj

    @property
    def cls(self):
        """Class (can be None) where the test function was collected."""
        if self.scope not in ("class", "function"):
            raise AttributeError(
                "cls not available in {}-scoped context".format(self.scope)
            )
        clscol = self._pyfuncitem.getparent(_pytest.python.Class)
        if clscol:
            return clscol.obj

    @property
    def instance(self):
        """Instance (can be None) on which test function was collected."""
        # unittest support hack, see _pytest.unittest.TestCaseFunction.
        try:
            return self._pyfuncitem._testcase
        except AttributeError:
            function = getattr(self, "function", None)
            return getattr(function, "__self__", None)

    @property
    def module(self):
        """Python module object where the test function was collected."""
        if self.scope not in ("function", "class", "module"):
            raise AttributeError(
                "module not available in {}-scoped context".format(self.scope)
            )
        return self._pyfuncitem.getparent(_pytest.python.Module).obj

    @property
    def fspath(self) -> py.path.local:
        """The file system path of the test module which collected this test."""
        if self.scope not in ("function", "class", "module", "package"):
            raise AttributeError(
                "module not available in {}-scoped context".format(self.scope)
            )
        # TODO: Remove ignore once _pyfuncitem is properly typed.
        return self._pyfuncitem.fspath  # type: ignore

    @property
    def keywords(self):
        """Keywords/markers dictionary for the underlying node."""
        return self.node.keywords

    @property
    def session(self):
        """Pytest session object."""
        return self._pyfuncitem.session

    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        """Add finalizer/teardown function to be called after the last test
        within the requesting test context finished execution."""
        # XXX usually this method is shadowed by fixturedef specific ones.
        self._addfinalizer(finalizer, scope=self.scope)

    def _addfinalizer(self, finalizer: Callable[[], object], scope) -> None:
        colitem = self._getscopeitem(scope)
        self._pyfuncitem.session._setupstate.addfinalizer(
            finalizer=finalizer, colitem=colitem
        )

    def applymarker(self, marker) -> None:
        """Apply a marker to a single test function invocation.

        This method is useful if you don't want to have a keyword/marker
        on all function invocations.

        :param marker:
            A :py:class:`_pytest.mark.MarkDecorator` object created by a call
            to ``pytest.mark.NAME(...)``.
        """
        self.node.add_marker(marker)

    def raiseerror(self, msg: Optional[str]) -> "NoReturn":
        """Raise a FixtureLookupError with the given message."""
        raise self._fixturemanager.FixtureLookupError(None, self, msg)

    def _fillfixtures(self) -> None:
        item = self._pyfuncitem
        fixturenames = getattr(item, "fixturenames", self.fixturenames)
        for argname in fixturenames:
            if argname not in item.funcargs:
                item.funcargs[argname] = self.getfixturevalue(argname)

    def getfixturevalue(self, argname: str) -> Any:
        """Dynamically run a named fixture function.

        Declaring fixtures via function argument is recommended where possible.
        But if you can only decide whether to use another fixture at test
        setup time, you may use this function to retrieve it inside a fixture
        or test function body.

        :raises pytest.FixtureLookupError:
            If the given fixture could not be found.
        """
        fixturedef = self._get_active_fixturedef(argname)
        assert fixturedef.cached_result is not None
        return fixturedef.cached_result[0]

    def _get_active_fixturedef(
        self, argname: str
    ) -> Union["FixtureDef[object]", PseudoFixtureDef[object]]:
        try:
            return self._fixture_defs[argname]
        except KeyError:
            try:
                fixturedef = self._getnextfixturedef(argname)
            except FixtureLookupError:
                if argname == "request":
                    cached_result = (self, [0], None)
                    scope = "function"  # type: _Scope
                    return PseudoFixtureDef(cached_result, scope)
                raise
        # Remove indent to prevent the python3 exception
        # from leaking into the call.
        self._compute_fixture_value(fixturedef)
        self._fixture_defs[argname] = fixturedef
        return fixturedef

    def _get_fixturestack(self) -> List["FixtureDef[Any]"]:
        current = self
        values = []  # type: List[FixtureDef[Any]]
        while 1:
            fixturedef = getattr(current, "_fixturedef", None)
            if fixturedef is None:
                values.reverse()
                return values
            values.append(fixturedef)
            assert isinstance(current, SubRequest)
            current = current._parent_request

    def _compute_fixture_value(self, fixturedef: "FixtureDef[object]") -> None:
        """Create a SubRequest based on "self" and call the execute method
        of the given FixtureDef object.

        This will force the FixtureDef object to throw away any previous
        results and compute a new fixture value, which will be stored into
        the FixtureDef object itself.
        """
        # prepare a subrequest object before calling fixture function
        # (latter managed by fixturedef)
        argname = fixturedef.argname
        funcitem = self._pyfuncitem
        scope = fixturedef.scope
        try:
            param = funcitem.callspec.getparam(argname)
        except (AttributeError, ValueError):
            param = NOTSET
            param_index = 0
            has_params = fixturedef.params is not None
            fixtures_not_supported = getattr(funcitem, "nofuncargs", False)
            if has_params and fixtures_not_supported:
                msg = (
                    "{name} does not support fixtures, maybe unittest.TestCase subclass?\n"
                    "Node id: {nodeid}\n"
                    "Function type: {typename}"
                ).format(
                    name=funcitem.name,
                    nodeid=funcitem.nodeid,
                    typename=type(funcitem).__name__,
                )
                fail(msg, pytrace=False)
            if has_params:
                frame = inspect.stack()[3]
                frameinfo = inspect.getframeinfo(frame[0])
                source_path = py.path.local(frameinfo.filename)
                source_lineno = frameinfo.lineno
                rel_source_path = source_path.relto(funcitem.config.rootdir)
                if rel_source_path:
                    source_path_str = rel_source_path
                else:
                    source_path_str = str(source_path)
                msg = (
                    "The requested fixture has no parameter defined for test:\n"
                    "    {}\n\n"
                    "Requested fixture '{}' defined in:\n{}"
                    "\n\nRequested here:\n{}:{}".format(
                        funcitem.nodeid,
                        fixturedef.argname,
                        getlocation(fixturedef.func, funcitem.config.rootdir),
                        source_path_str,
                        source_lineno,
                    )
                )
                fail(msg, pytrace=False)
        else:
            param_index = funcitem.callspec.indices[argname]
            # If a parametrize invocation set a scope it will override
            # the static scope defined with the fixture function.
            paramscopenum = funcitem.callspec._arg2scopenum.get(argname)
            if paramscopenum is not None:
                scope = scopes[paramscopenum]

        subrequest = SubRequest(self, scope, param, param_index, fixturedef)

        # Check if a higher-level scoped fixture accesses a lower level one.
        subrequest._check_scope(argname, self.scope, scope)
        try:
            # Call the fixture function.
            fixturedef.execute(request=subrequest)
        finally:
            self._schedule_finalizers(fixturedef, subrequest)

    def _schedule_finalizers(
        self, fixturedef: "FixtureDef[object]", subrequest: "SubRequest"
    ) -> None:
        # If fixture function failed it might have registered finalizers.
        self.session._setupstate.addfinalizer(
            functools.partial(fixturedef.finish, request=subrequest), subrequest.node
        )

    def _check_scope(self, argname, invoking_scope: "_Scope", requested_scope) -> None:
        if argname == "request":
            return
        if scopemismatch(invoking_scope, requested_scope):
            # Try to report something helpful.
            lines = self._factorytraceback()
            fail(
                "ScopeMismatch: You tried to access the %r scoped "
                "fixture %r with a %r scoped request object, "
                "involved factories\n%s"
                % ((requested_scope, argname, invoking_scope, "\n".join(lines))),
                pytrace=False,
            )

    def _factorytraceback(self) -> List[str]:
        lines = []
        for fixturedef in self._get_fixturestack():
            factory = fixturedef.func
            fs, lineno = getfslineno(factory)
            p = self._pyfuncitem.session.fspath.bestrelpath(fs)
            args = _format_args(factory)
            lines.append("%s:%d:  def %s%s" % (p, lineno + 1, factory.__name__, args))
        return lines

    def _getscopeitem(self, scope):
        if scope == "function":
            # This might also be a non-function Item despite its attribute name.
            return self._pyfuncitem
        if scope == "package":
            # FIXME: _fixturedef is not defined on FixtureRequest (this class),
            # but on FixtureRequest (a subclass).
            node = get_scope_package(self._pyfuncitem, self._fixturedef)  # type: ignore[attr-defined]
        else:
            node = get_scope_node(self._pyfuncitem, scope)
        if node is None and scope == "class":
            # Fallback to function item itself.
            node = self._pyfuncitem
        assert node, 'Could not obtain a node for scope "{}" for function {!r}'.format(
            scope, self._pyfuncitem
        )
        return node

    def __repr__(self) -> str:
        return "<FixtureRequest for %r>" % (self.node)


@final
class SubRequest(FixtureRequest):
    """A sub request for handling getting a fixture from a test function/fixture."""

    def __init__(
        self,
        request: "FixtureRequest",
        scope: "_Scope",
        param,
        param_index: int,
        fixturedef: "FixtureDef[object]",
    ) -> None:
        self._parent_request = request
        self.fixturename = fixturedef.argname
        if param is not NOTSET:
            self.param = param
        self.param_index = param_index
        self.scope = scope
        self._fixturedef = fixturedef
        self._pyfuncitem = request._pyfuncitem
        self._fixture_defs = request._fixture_defs
        self._arg2fixturedefs = request._arg2fixturedefs
        self._arg2index = request._arg2index
        self._fixturemanager = request._fixturemanager

    def __repr__(self) -> str:
        return "<SubRequest {!r} for {!r}>".format(self.fixturename, self._pyfuncitem)

    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        self._fixturedef.addfinalizer(finalizer)

    def _schedule_finalizers(
        self, fixturedef: "FixtureDef[object]", subrequest: "SubRequest"
    ) -> None:
        # If the executing fixturedef was not explicitly requested in the argument list (via
        # getfixturevalue inside the fixture call) then ensure this fixture def will be finished
        # first.
        if fixturedef.argname not in self.fixturenames:
            fixturedef.addfinalizer(
                functools.partial(self._fixturedef.finish, request=self)
            )
        super()._schedule_finalizers(fixturedef, subrequest)


scopes = ["session", "package", "module", "class", "function"]  # type: List[_Scope]
scopenum_function = scopes.index("function")


def scopemismatch(currentscope: "_Scope", newscope: "_Scope") -> bool:
    return scopes.index(newscope) > scopes.index(currentscope)


def scope2index(scope: str, descr: str, where: Optional[str] = None) -> int:
    """Look up the index of ``scope`` and raise a descriptive value error
    if not defined."""
    strscopes = scopes  # type: Sequence[str]
    try:
        return strscopes.index(scope)
    except ValueError:
        fail(
            "{} {}got an unexpected scope value '{}'".format(
                descr, "from {} ".format(where) if where else "", scope
            ),
            pytrace=False,
        )


@final
class FixtureLookupError(LookupError):
    """Could not return a requested fixture (missing or invalid)."""

    def __init__(
        self, argname: Optional[str], request: FixtureRequest, msg: Optional[str] = None
    ) -> None:
        self.argname = argname
        self.request = request
        self.fixturestack = request._get_fixturestack()
        self.msg = msg

    def formatrepr(self) -> "FixtureLookupErrorRepr":
        tblines = []  # type: List[str]
        addline = tblines.append
        stack = [self.request._pyfuncitem.obj]
        stack.extend(map(lambda x: x.func, self.fixturestack))
        msg = self.msg
        if msg is not None:
            # The last fixture raise an error, let's present
            # it at the requesting side.
            stack = stack[:-1]
        for function in stack:
            fspath, lineno = getfslineno(function)
            try:
                lines, _ = inspect.getsourcelines(get_real_func(function))
            except (OSError, IndexError, TypeError):
                error_msg = "file %s, line %s: source code not available"
                addline(error_msg % (fspath, lineno + 1))
            else:
                addline("file {}, line {}".format(fspath, lineno + 1))
                for i, line in enumerate(lines):
                    line = line.rstrip()
                    addline("  " + line)
                    if line.lstrip().startswith("def"):
                        break

        if msg is None:
            fm = self.request._fixturemanager
            available = set()
            parentid = self.request._pyfuncitem.parent.nodeid
            for name, fixturedefs in fm._arg2fixturedefs.items():
                faclist = list(fm._matchfactories(fixturedefs, parentid))
                if faclist:
                    available.add(name)
            if self.argname in available:
                msg = " recursive dependency involving fixture '{}' detected".format(
                    self.argname
                )
            else:
                msg = "fixture '{}' not found".format(self.argname)
            msg += "\n available fixtures: {}".format(", ".join(sorted(available)))
            msg += "\n use 'pytest --fixtures [testpath]' for help on them."

        return FixtureLookupErrorRepr(fspath, lineno, tblines, msg, self.argname)


class FixtureLookupErrorRepr(TerminalRepr):
    def __init__(
        self,
        filename: Union[str, py.path.local],
        firstlineno: int,
        tblines: Sequence[str],
        errorstring: str,
        argname: Optional[str],
    ) -> None:
        self.tblines = tblines
        self.errorstring = errorstring
        self.filename = filename
        self.firstlineno = firstlineno
        self.argname = argname

    def toterminal(self, tw: TerminalWriter) -> None:
        # tw.line("FixtureLookupError: %s" %(self.argname), red=True)
        for tbline in self.tblines:
            tw.line(tbline.rstrip())
        lines = self.errorstring.split("\n")
        if lines:
            tw.line(
                "{}       {}".format(FormattedExcinfo.fail_marker, lines[0].strip()),
                red=True,
            )
            for line in lines[1:]:
                tw.line(
                    "{}       {}".format(FormattedExcinfo.flow_marker, line.strip()),
                    red=True,
                )
        tw.line()
        tw.line("%s:%d" % (self.filename, self.firstlineno + 1))


def fail_fixturefunc(fixturefunc, msg: str) -> "NoReturn":
    fs, lineno = getfslineno(fixturefunc)
    location = "{}:{}".format(fs, lineno + 1)
    source = _pytest._code.Source(fixturefunc)
    fail(msg + ":\n\n" + str(source.indent()) + "\n" + location, pytrace=False)


def call_fixture_func(
    fixturefunc: "_FixtureFunc[_FixtureValue]", request: FixtureRequest, kwargs
) -> _FixtureValue:
    if is_generator(fixturefunc):
        fixturefunc = cast(
            Callable[..., Generator[_FixtureValue, None, None]], fixturefunc
        )
        generator = fixturefunc(**kwargs)
        try:
            fixture_result = next(generator)
        except StopIteration:
            raise ValueError(
                "{} did not yield a value".format(request.fixturename)
            ) from None
        finalizer = functools.partial(_teardown_yield_fixture, fixturefunc, generator)
        request.addfinalizer(finalizer)
    else:
        fixturefunc = cast(Callable[..., _FixtureValue], fixturefunc)
        fixture_result = fixturefunc(**kwargs)
    return fixture_result


def _teardown_yield_fixture(fixturefunc, it) -> None:
    """Execute the teardown of a fixture function by advancing the iterator
    after the yield and ensure the iteration ends (if not it means there is
    more than one yield in the function)."""
    try:
        next(it)
    except StopIteration:
        pass
    else:
        fail_fixturefunc(fixturefunc, "fixture function has more than one 'yield'")


def _eval_scope_callable(
    scope_callable: "Callable[[str, Config], _Scope]",
    fixture_name: str,
    config: Config,
) -> "_Scope":
    try:
        # Type ignored because there is no typing mechanism to specify
        # keyword arguments, currently.
        result = scope_callable(fixture_name=fixture_name, config=config)  # type: ignore[call-arg]
    except Exception as e:
        raise TypeError(
            "Error evaluating {} while defining fixture '{}'.\n"
            "Expected a function with the signature (*, fixture_name, config)".format(
                scope_callable, fixture_name
            )
        ) from e
    if not isinstance(result, str):
        fail(
            "Expected {} to return a 'str' while defining fixture '{}', but it returned:\n"
            "{!r}".format(scope_callable, fixture_name, result),
            pytrace=False,
        )
    return result


@final
class FixtureDef(Generic[_FixtureValue]):
    """A container for a factory definition."""

    def __init__(
        self,
        fixturemanager: "FixtureManager",
        baseid,
        argname: str,
        func: "_FixtureFunc[_FixtureValue]",
        scope: "Union[_Scope, Callable[[str, Config], _Scope]]",
        params: Optional[Sequence[object]],
        unittest: bool = False,
        ids: Optional[
            Union[
                Tuple[Union[None, str, float, int, bool], ...],
                Callable[[Any], Optional[object]],
            ]
        ] = None,
    ) -> None:
        self._fixturemanager = fixturemanager
        self.baseid = baseid or ""
        self.has_location = baseid is not None
        self.func = func
        self.argname = argname
        if callable(scope):
            scope_ = _eval_scope_callable(scope, argname, fixturemanager.config)
        else:
            scope_ = scope
        self.scopenum = scope2index(
            # TODO: Check if the `or` here is really necessary.
            scope_ or "function",  # type: ignore[unreachable]
            descr="Fixture '{}'".format(func.__name__),
            where=baseid,
        )
        self.scope = scope_
        self.params = params  # type: Optional[Sequence[object]]
        self.argnames = getfuncargnames(
            func, name=argname, is_method=unittest
        )  # type: Tuple[str, ...]
        self.unittest = unittest
        self.ids = ids
        self.cached_result = None  # type: Optional[_FixtureCachedResult[_FixtureValue]]
        self._finalizers = []  # type: List[Callable[[], object]]

    def addfinalizer(self, finalizer: Callable[[], object]) -> None:
        self._finalizers.append(finalizer)

    def finish(self, request: SubRequest) -> None:
        exc = None
        try:
            while self._finalizers:
                try:
                    func = self._finalizers.pop()
                    func()
                except BaseException as e:
                    # XXX Only first exception will be seen by user,
                    #     ideally all should be reported.
                    if exc is None:
                        exc = e
            if exc:
                raise exc
        finally:
            hook = self._fixturemanager.session.gethookproxy(request.node.fspath)
            hook.pytest_fixture_post_finalizer(fixturedef=self, request=request)
            # Even if finalization fails, we invalidate the cached fixture
            # value and remove all finalizers because they may be bound methods
            # which will keep instances alive.
            self.cached_result = None
            self._finalizers = []

    def execute(self, request: SubRequest) -> _FixtureValue:
        # Get required arguments and register our own finish()
        # with their finalization.
        for argname in self.argnames:
            fixturedef = request._get_active_fixturedef(argname)
            if argname != "request":
                # PseudoFixtureDef is only for "request".
                assert isinstance(fixturedef, FixtureDef)
                fixturedef.addfinalizer(functools.partial(self.finish, request=request))

        my_cache_key = self.cache_key(request)
        if self.cached_result is not None:
            # note: comparison with `==` can fail (or be expensive) for e.g.
            # numpy arrays (#6497).
            cache_key = self.cached_result[1]
            if my_cache_key is cache_key:
                if self.cached_result[2] is not None:
                    _, val, tb = self.cached_result[2]
                    raise val.with_traceback(tb)
                else:
                    result = self.cached_result[0]
                    return result
            # We have a previous but differently parametrized fixture instance
            # so we need to tear it down before creating a new one.
            self.finish(request)
            assert self.cached_result is None

        hook = self._fixturemanager.session.gethookproxy(request.node.fspath)
        result = hook.pytest_fixture_setup(fixturedef=self, request=request)
        return result

    def cache_key(self, request: SubRequest) -> object:
        return request.param_index if not hasattr(request, "param") else request.param

    def __repr__(self) -> str:
        return "<FixtureDef argname={!r} scope={!r} baseid={!r}>".format(
            self.argname, self.scope, self.baseid
        )


def resolve_fixture_function(
    fixturedef: FixtureDef[_FixtureValue], request: FixtureRequest
) -> "_FixtureFunc[_FixtureValue]":
    """Get the actual callable that can be called to obtain the fixture
    value, dealing with unittest-specific instances and bound methods."""
    fixturefunc = fixturedef.func
    if fixturedef.unittest:
        if request.instance is not None:
            # Bind the unbound method to the TestCase instance.
            fixturefunc = fixturedef.func.__get__(request.instance)  # type: ignore[union-attr]
    else:
        # The fixture function needs to be bound to the actual
        # request.instance so that code working with "fixturedef" behaves
        # as expected.
        if request.instance is not None:
            # Handle the case where fixture is defined not in a test class, but some other class
            # (for example a plugin class with a fixture), see #2270.
            if hasattr(fixturefunc, "__self__") and not isinstance(
                request.instance, fixturefunc.__self__.__class__  # type: ignore[union-attr]
            ):
                return fixturefunc
            fixturefunc = getimfunc(fixturedef.func)
            if fixturefunc != fixturedef.func:
                fixturefunc = fixturefunc.__get__(request.instance)  # type: ignore[union-attr]
    return fixturefunc


def pytest_fixture_setup(
    fixturedef: FixtureDef[_FixtureValue], request: SubRequest
) -> _FixtureValue:
    """Execution of fixture setup."""
    kwargs = {}
    for argname in fixturedef.argnames:
        fixdef = request._get_active_fixturedef(argname)
        assert fixdef.cached_result is not None
        result, arg_cache_key, exc = fixdef.cached_result
        request._check_scope(argname, request.scope, fixdef.scope)
        kwargs[argname] = result

    fixturefunc = resolve_fixture_function(fixturedef, request)
    my_cache_key = fixturedef.cache_key(request)
    try:
        result = call_fixture_func(fixturefunc, request, kwargs)
    except TEST_OUTCOME:
        exc_info = sys.exc_info()
        assert exc_info[0] is not None
        fixturedef.cached_result = (None, my_cache_key, exc_info)
        raise
    fixturedef.cached_result = (result, my_cache_key, None)
    return result


def _ensure_immutable_ids(
    ids: Optional[
        Union[
            Iterable[Union[None, str, float, int, bool]],
            Callable[[Any], Optional[object]],
        ]
    ],
) -> Optional[
    Union[
        Tuple[Union[None, str, float, int, bool], ...],
        Callable[[Any], Optional[object]],
    ]
]:
    if ids is None:
        return None
    if callable(ids):
        return ids
    return tuple(ids)


def _params_converter(
    params: Optional[Iterable[object]],
) -> Optional[Tuple[object, ...]]:
    return tuple(params) if params is not None else None


def wrap_function_to_error_out_if_called_directly(function, fixture_marker):
    """Wrap the given fixture function so we can raise an error about it being called directly,
    instead of used as an argument in a test function."""
    message = (
        'Fixture "{name}" called directly. Fixtures are not meant to be called directly,\n'
        "but are created automatically when test functions request them as parameters.\n"
        "See https://docs.pytest.org/en/stable/fixture.html for more information about fixtures, and\n"
        "https://docs.pytest.org/en/stable/deprecations.html#calling-fixtures-directly about how to update your code."
    ).format(name=fixture_marker.name or function.__name__)

    @functools.wraps(function)
    def result(*args, **kwargs):
        fail(message, pytrace=False)

    # Keep reference to the original function in our own custom attribute so we don't unwrap
    # further than this point and lose useful wrappings like @mock.patch (#3774).
    result.__pytest_wrapped__ = _PytestWrapper(function)  # type: ignore[attr-defined]

    return result


@final
@attr.s(frozen=True)
class FixtureFunctionMarker:
    scope = attr.ib(type="Union[_Scope, Callable[[str, Config], _Scope]]")
    params = attr.ib(type=Optional[Tuple[object, ...]], converter=_params_converter)
    autouse = attr.ib(type=bool, default=False)
    ids = attr.ib(
        type=Union[
            Tuple[Union[None, str, float, int, bool], ...],
            Callable[[Any], Optional[object]],
        ],
        default=None,
        converter=_ensure_immutable_ids,
    )
    name = attr.ib(type=Optional[str], default=None)

    def __call__(self, function: _FixtureFunction) -> _FixtureFunction:
        if inspect.isclass(function):
            raise ValueError("class fixtures not supported (maybe in the future)")

        if getattr(function, "_pytestfixturefunction", False):
            raise ValueError(
                "fixture is being applied more than once to the same function"
            )

        function = wrap_function_to_error_out_if_called_directly(function, self)

        name = self.name or function.__name__
        if name == "request":
            location = getlocation(function)
            fail(
                "'request' is a reserved word for fixtures, use another name:\n  {}".format(
                    location
                ),
                pytrace=False,
            )

        # Type ignored because https://github.com/python/mypy/issues/2087.
        function._pytestfixturefunction = self  # type: ignore[attr-defined]
        return function


@overload
def fixture(
    fixture_function: _FixtureFunction,
    *,
    scope: "Union[_Scope, Callable[[str, Config], _Scope]]" = ...,
    params: Optional[Iterable[object]] = ...,
    autouse: bool = ...,
    ids: Optional[
        Union[
            Iterable[Union[None, str, float, int, bool]],
            Callable[[Any], Optional[object]],
        ]
    ] = ...,
    name: Optional[str] = ...
) -> _FixtureFunction:
    ...


@overload  # noqa: F811
def fixture(  # noqa: F811
    fixture_function: None = ...,
    *,
    scope: "Union[_Scope, Callable[[str, Config], _Scope]]" = ...,
    params: Optional[Iterable[object]] = ...,
    autouse: bool = ...,
    ids: Optional[
        Union[
            Iterable[Union[None, str, float, int, bool]],
            Callable[[Any], Optional[object]],
        ]
    ] = ...,
    name: Optional[str] = None
) -> FixtureFunctionMarker:
    ...


def fixture(  # noqa: F811
    fixture_function: Optional[_FixtureFunction] = None,
    *,
    scope: "Union[_Scope, Callable[[str, Config], _Scope]]" = "function",
    params: Optional[Iterable[object]] = None,
    autouse: bool = False,
    ids: Optional[
        Union[
            Iterable[Union[None, str, float, int, bool]],
            Callable[[Any], Optional[object]],
        ]
    ] = None,
    name: Optional[str] = None
) -> Union[FixtureFunctionMarker, _FixtureFunction]:
    """Decorator to mark a fixture factory function.

    This decorator can be used, with or without parameters, to define a
    fixture function.

    The name of the fixture function can later be referenced to cause its
    invocation ahead of running tests: test modules or classes can use the
    ``pytest.mark.usefixtures(fixturename)`` marker.

    Test functions can directly use fixture names as input arguments in which
    case the fixture instance returned from the fixture function will be
    injected.

    Fixtures can provide their values to test functions using ``return`` or
    ``yield`` statements. When using ``yield`` the code block after the
    ``yield`` statement is executed as teardown code regardless of the test
    outcome, and must yield exactly once.

    :param scope:
        The scope for which this fixture is shared; one of ``"function"``
        (default), ``"class"``, ``"module"``, ``"package"`` or ``"session"``.

        This parameter may also be a callable which receives ``(fixture_name, config)``
        as parameters, and must return a ``str`` with one of the values mentioned above.

        See :ref:`dynamic scope` in the docs for more information.

    :param params:
        An optional list of parameters which will cause multiple invocations
        of the fixture function and all of the tests using it. The current
        parameter is available in ``request.param``.

    :param autouse:
        If True, the fixture func is activated for all tests that can see it.
        If False (the default), an explicit reference is needed to activate
        the fixture.

    :param ids:
        List of string ids each corresponding to the params so that they are
        part of the test id. If no ids are provided they will be generated
        automatically from the params.

    :param name:
        The name of the fixture. This defaults to the name of the decorated
        function. If a fixture is used in the same module in which it is
        defined, the function name of the fixture will be shadowed by the
        function arg that requests the fixture; one way to resolve this is to
        name the decorated function ``fixture_<fixturename>`` and then use
        ``@pytest.fixture(name='<fixturename>')``.
    """
    fixture_marker = FixtureFunctionMarker(
        scope=scope, params=params, autouse=autouse, ids=ids, name=name,
    )

    # Direct decoration.
    if fixture_function:
        return fixture_marker(fixture_function)

    return fixture_marker


def yield_fixture(
    fixture_function=None,
    *args,
    scope="function",
    params=None,
    autouse=False,
    ids=None,
    name=None
):
    """(Return a) decorator to mark a yield-fixture factory function.

    .. deprecated:: 3.0
        Use :py:func:`pytest.fixture` directly instead.
    """
    return fixture(
        fixture_function,
        *args,
        scope=scope,
        params=params,
        autouse=autouse,
        ids=ids,
        name=name,
    )


@fixture(scope="session")
def pytestconfig(request: FixtureRequest) -> Config:
    """Session-scoped fixture that returns the :class:`_pytest.config.Config` object.

    Example::

        def test_foo(pytestconfig):
            if pytestconfig.getoption("verbose") > 0:
                ...

    """
    return request.config


def pytest_addoption(parser: Parser) -> None:
    parser.addini(
        "usefixtures",
        type="args",
        default=[],
        help="list of default fixtures to be used with this project",
    )


class FixtureManager:
    """pytest fixture definitions and information is stored and managed
    from this class.

    During collection fm.parsefactories() is called multiple times to parse
    fixture function definitions into FixtureDef objects and internal
    data structures.

    During collection of test functions, metafunc-mechanics instantiate
    a FuncFixtureInfo object which is cached per node/func-name.
    This FuncFixtureInfo object is later retrieved by Function nodes
    which themselves offer a fixturenames attribute.

    The FuncFixtureInfo object holds information about fixtures and FixtureDefs
    relevant for a particular function. An initial list of fixtures is
    assembled like this:

    - ini-defined usefixtures
    - autouse-marked fixtures along the collection chain up from the function
    - usefixtures markers at module/class/function level
    - test function funcargs

    Subsequently the funcfixtureinfo.fixturenames attribute is computed
    as the closure of the fixtures needed to setup the initial fixtures,
    i.e. fixtures needed by fixture functions themselves are appended
    to the fixturenames list.

    Upon the test-setup phases all fixturenames are instantiated, retrieved
    by a lookup of their FuncFixtureInfo.
    """

    FixtureLookupError = FixtureLookupError
    FixtureLookupErrorRepr = FixtureLookupErrorRepr

    def __init__(self, session: "Session") -> None:
        self.session = session
        self.config = session.config  # type: Config
        self._arg2fixturedefs = {}  # type: Dict[str, List[FixtureDef[Any]]]
        self._holderobjseen = set()  # type: Set[object]
        self._nodeid_and_autousenames = [
            ("", self.config.getini("usefixtures"))
        ]  # type: List[Tuple[str, List[str]]]
        session.config.pluginmanager.register(self, "funcmanage")

    def _get_direct_parametrize_args(self, node: "nodes.Node") -> List[str]:
        """Return all direct parametrization arguments of a node, so we don't
        mistake them for fixtures.

        Check https://github.com/pytest-dev/pytest/issues/5036.

        These things are done later as well when dealing with parametrization
        so this could be improved.
        """
        parametrize_argnames = []  # type: List[str]
        for marker in node.iter_markers(name="parametrize"):
            if not marker.kwargs.get("indirect", False):
                p_argnames, _ = ParameterSet._parse_parametrize_args(
                    *marker.args, **marker.kwargs
                )
                parametrize_argnames.extend(p_argnames)

        return parametrize_argnames

    def getfixtureinfo(
        self, node: "nodes.Node", func, cls, funcargs: bool = True
    ) -> FuncFixtureInfo:
        if funcargs and not getattr(node, "nofuncargs", False):
            argnames = getfuncargnames(func, name=node.name, cls=cls)
        else:
            argnames = ()

        usefixtures = tuple(
            arg for mark in node.iter_markers(name="usefixtures") for arg in mark.args
        )
        initialnames = usefixtures + argnames
        fm = node.session._fixturemanager
        initialnames, names_closure, arg2fixturedefs = fm.getfixtureclosure(
            initialnames, node, ignore_args=self._get_direct_parametrize_args(node)
        )
        return FuncFixtureInfo(argnames, initialnames, names_closure, arg2fixturedefs)

    def pytest_plugin_registered(self, plugin: _PluggyPlugin) -> None:
        nodeid = None
        try:
            p = absolutepath(plugin.__file__)  # type: ignore[attr-defined]
        except AttributeError:
            pass
        else:
            from _pytest import nodes

            # Construct the base nodeid which is later used to check
            # what fixtures are visible for particular tests (as denoted
            # by their test id).
            if p.name.startswith("conftest.py"):
                try:
                    nodeid = str(p.parent.relative_to(self.config.rootpath))
                except ValueError:
                    nodeid = ""
                if nodeid == ".":
                    nodeid = ""
                if os.sep != nodes.SEP:
                    nodeid = nodeid.replace(os.sep, nodes.SEP)

        self.parsefactories(plugin, nodeid)

    def _getautousenames(self, nodeid: str) -> List[str]:
        """Return a list of fixture names to be used."""
        autousenames = []  # type: List[str]
        for baseid, basenames in self._nodeid_and_autousenames:
            if nodeid.startswith(baseid):
                if baseid:
                    i = len(baseid)
                    nextchar = nodeid[i : i + 1]
                    if nextchar and nextchar not in ":/":
                        continue
                autousenames.extend(basenames)
        return autousenames

    def getfixtureclosure(
        self, fixturenames: Tuple[str, ...], parentnode, ignore_args: Sequence[str] = ()
    ) -> Tuple[Tuple[str, ...], List[str], Dict[str, Sequence[FixtureDef[Any]]]]:
        # Collect the closure of all fixtures, starting with the given
        # fixturenames as the initial set.  As we have to visit all
        # factory definitions anyway, we also return an arg2fixturedefs
        # mapping so that the caller can reuse it and does not have
        # to re-discover fixturedefs again for each fixturename
        # (discovering matching fixtures for a given name/node is expensive).

        parentid = parentnode.nodeid
        fixturenames_closure = self._getautousenames(parentid)

        def merge(otherlist: Iterable[str]) -> None:
            for arg in otherlist:
                if arg not in fixturenames_closure:
                    fixturenames_closure.append(arg)

        merge(fixturenames)

        # At this point, fixturenames_closure contains what we call "initialnames",
        # which is a set of fixturenames the function immediately requests. We
        # need to return it as well, so save this.
        initialnames = tuple(fixturenames_closure)

        arg2fixturedefs = {}  # type: Dict[str, Sequence[FixtureDef[Any]]]
        lastlen = -1
        while lastlen != len(fixturenames_closure):
            lastlen = len(fixturenames_closure)
            for argname in fixturenames_closure:
                if argname in ignore_args:
                    continue
                if argname in arg2fixturedefs:
                    continue
                fixturedefs = self.getfixturedefs(argname, parentid)
                if fixturedefs:
                    arg2fixturedefs[argname] = fixturedefs
                    merge(fixturedefs[-1].argnames)

        def sort_by_scope(arg_name: str) -> int:
            try:
                fixturedefs = arg2fixturedefs[arg_name]
            except KeyError:
                return scopes.index("function")
            else:
                return fixturedefs[-1].scopenum

        fixturenames_closure.sort(key=sort_by_scope)
        return initialnames, fixturenames_closure, arg2fixturedefs

    def pytest_generate_tests(self, metafunc: "Metafunc") -> None:
        """Generate new tests based on parametrized fixtures used by the given metafunc"""

        def get_parametrize_mark_argnames(mark: Mark) -> Sequence[str]:
            args, _ = ParameterSet._parse_parametrize_args(*mark.args, **mark.kwargs)
            return args

        for argname in metafunc.fixturenames:
            # Get the FixtureDefs for the argname.
            fixture_defs = metafunc._arg2fixturedefs.get(argname)
            if not fixture_defs:
                # Will raise FixtureLookupError at setup time if not parametrized somewhere
                # else (e.g @pytest.mark.parametrize)
                continue

            # If the test itself parametrizes using this argname, give it
            # precedence.
            if any(
                argname in get_parametrize_mark_argnames(mark)
                for mark in metafunc.definition.iter_markers("parametrize")
            ):
                continue

            # In the common case we only look at the fixture def with the
            # closest scope (last in the list). But if the fixture overrides
            # another fixture, while requesting the super fixture, keep going
            # in case the super fixture is parametrized (#1953).
            for fixturedef in reversed(fixture_defs):
                # Fixture is parametrized, apply it and stop.
                if fixturedef.params is not None:
                    metafunc.parametrize(
                        argname,
                        fixturedef.params,
                        indirect=True,
                        scope=fixturedef.scope,
                        ids=fixturedef.ids,
                    )
                    break

                # Not requesting the overridden super fixture, stop.
                if argname not in fixturedef.argnames:
                    break

                # Try next super fixture, if any.

    def pytest_collection_modifyitems(self, items: "List[nodes.Item]") -> None:
        # Separate parametrized setups.
        items[:] = reorder_items(items)

    def parsefactories(
        self, node_or_obj, nodeid=NOTSET, unittest: bool = False
    ) -> None:
        if nodeid is not NOTSET:
            holderobj = node_or_obj
        else:
            holderobj = node_or_obj.obj
            nodeid = node_or_obj.nodeid
        if holderobj in self._holderobjseen:
            return

        self._holderobjseen.add(holderobj)
        autousenames = []
        for name in dir(holderobj):
            # The attribute can be an arbitrary descriptor, so the attribute
            # access below can raise. safe_getatt() ignores such exceptions.
            obj = safe_getattr(holderobj, name, None)
            marker = getfixturemarker(obj)
            if not isinstance(marker, FixtureFunctionMarker):
                # Magic globals  with __getattr__ might have got us a wrong
                # fixture attribute.
                continue

            if marker.name:
                name = marker.name

            # During fixture definition we wrap the original fixture function
            # to issue a warning if called directly, so here we unwrap it in
            # order to not emit the warning when pytest itself calls the
            # fixture function.
            obj = get_real_method(obj, holderobj)

            fixture_def = FixtureDef(
                fixturemanager=self,
                baseid=nodeid,
                argname=name,
                func=obj,
                scope=marker.scope,
                params=marker.params,
                unittest=unittest,
                ids=marker.ids,
            )

            faclist = self._arg2fixturedefs.setdefault(name, [])
            if fixture_def.has_location:
                faclist.append(fixture_def)
            else:
                # fixturedefs with no location are at the front
                # so this inserts the current fixturedef after the
                # existing fixturedefs from external plugins but
                # before the fixturedefs provided in conftests.
                i = len([f for f in faclist if not f.has_location])
                faclist.insert(i, fixture_def)
            if marker.autouse:
                autousenames.append(name)

        if autousenames:
            self._nodeid_and_autousenames.append((nodeid or "", autousenames))

    def getfixturedefs(
        self, argname: str, nodeid: str
    ) -> Optional[Sequence[FixtureDef[Any]]]:
        """Get a list of fixtures which are applicable to the given node id.

        :param str argname: Name of the fixture to search for.
        :param str nodeid: Full node id of the requesting test.
        :rtype: Sequence[FixtureDef]
        """
        try:
            fixturedefs = self._arg2fixturedefs[argname]
        except KeyError:
            return None
        return tuple(self._matchfactories(fixturedefs, nodeid))

    def _matchfactories(
        self, fixturedefs: Iterable[FixtureDef[Any]], nodeid: str
    ) -> Iterator[FixtureDef[Any]]:
        from _pytest import nodes

        for fixturedef in fixturedefs:
            if nodes.ischildnode(fixturedef.baseid, nodeid):
                yield fixturedef



from .. import BaseProvider

localized = True


class Provider(BaseProvider):
    formats = (
        '{{last_name}} {{company_suffix}}',
        '{{last_name}}-{{last_name}}',
        '{{last_name}}, {{last_name}} and {{last_name}}',
    )

    company_suffixes = ('Inc', 'and Sons', 'LLC', 'Group', 'PLC', 'Ltd')

    catch_phrase_words = (
        ('Adaptive',
         'Advanced',
         'Ameliorated',
         'Assimilated',
         'Automated',
         'Balanced',
         'Business-focused',
         'Centralized',
         'Cloned',
         'Compatible',
         'Configurable',
         'Cross-group',
         'Cross-platform',
         'Customer-focused',
         'Customizable',
         'Decentralized',
         'De-engineered',
         'Devolved',
         'Digitized',
         'Distributed',
         'Diverse',
         'Down-sized',
         'Enhanced',
         'Enterprise-wide',
         'Ergonomic',
         'Exclusive',
         'Expanded',
         'Extended',
         'Face-to-face',
         'Focused',
         'Front-line',
         'Fully-configurable',
         'Function-based',
         'Fundamental',
         'Future-proofed',
         'Grass-roots',
         'Horizontal',
         'Implemented',
         'Innovative',
         'Integrated',
         'Intuitive',
         'Inverse',
         'Managed',
         'Mandatory',
         'Monitored',
         'Multi-channeled',
         'Multi-lateral',
         'Multi-layered',
         'Multi-tiered',
         'Networked',
         'Object-based',
         'Open-architected',
         'Open-source',
         'Operative',
         'Optimized',
         'Optional',
         'Organic',
         'Organized',
         'Persevering',
         'Persistent',
         'Phased',
         'Polarized',
         'Pre-emptive',
         'Proactive',
         'Profit-focused',
         'Profound',
         'Programmable',
         'Progressive',
         'Public-key',
         'Quality-focused',
         'Reactive',
         'Realigned',
         'Re-contextualized',
         'Re-engineered',
         'Reduced',
         'Reverse-engineered',
         'Right-sized',
         'Robust',
         'Seamless',
         'Secured',
         'Self-enabling',
         'Sharable',
         'Stand-alone',
         'Streamlined',
         'Switchable',
         'Synchronized',
         'Synergistic',
         'Synergized',
         'Team-oriented',
         'Total',
         'Triple-buffered',
         'Universal',
         'Up-sized',
         'Upgradable',
         'User-centric',
         'User-friendly',
         'Versatile',
         'Virtual',
         'Visionary',
         'Vision-oriented'),
        ('24hour',
         '24/7',
         '3rdgeneration',
         '4thgeneration',
         '5thgeneration',
         '6thgeneration',
         'actuating',
         'analyzing',
         'asymmetric',
         'asynchronous',
         'attitude-oriented',
         'background',
         'bandwidth-monitored',
         'bi-directional',
         'bifurcated',
         'bottom-line',
         'clear-thinking',
         'client-driven',
         'client-server',
         'coherent',
         'cohesive',
         'composite',
         'context-sensitive',
         'contextually-based',
         'content-based',
         'dedicated',
         'demand-driven',
         'didactic',
         'directional',
         'discrete',
         'disintermediate',
         'dynamic',
         'eco-centric',
         'empowering',
         'encompassing',
         'even-keeled',
         'executive',
         'explicit',
         'exuding',
         'fault-tolerant',
         'foreground',
         'fresh-thinking',
         'full-range',
         'global',
         'grid-enabled',
         'heuristic',
         'high-level',
         'holistic',
         'homogeneous',
         'human-resource',
         'hybrid',
         'impactful',
         'incremental',
         'intangible',
         'interactive',
         'intermediate',
         'leadingedge',
         'local',
         'logistical',
         'maximized',
         'methodical',
         'mission-critical',
         'mobile',
         'modular',
         'motivating',
         'multimedia',
         'multi-state',
         'multi-tasking',
         'national',
         'needs-based',
         'neutral',
         'next generation',
         'non-volatile',
         'object-oriented',
         'optimal',
         'optimizing',
         'radical',
         'real-time',
         'reciprocal',
         'regional',
         'responsive',
         'scalable',
         'secondary',
         'solution-oriented',
         'stable',
         'static',
         'systematic',
         'systemic',
         'system-worthy',
         'tangible',
         'tertiary',
         'transitional',
         'uniform',
         'upward-trending',
         'user-facing',
         'value-added',
         'web-enabled',
         'well-modulated',
         'zero administration',
         'zero-defect',
         'zero tolerance'),
        ('ability',
         'access',
         'adapter',
         'algorithm',
         'alliance',
         'analyzer',
         'application',
         'approach',
         'architecture',
         'archive',
         'artificial intelligence',
         'array',
         'attitude',
         'benchmark',
         'budgetary management',
         'capability',
         'capacity',
         'challenge',
         'circuit',
         'collaboration',
         'complexity',
         'concept',
         'conglomeration',
         'contingency',
         'core',
         'customer loyalty',
         'database',
         'data-warehouse',
         'definition',
         'emulation',
         'encoding',
         'encryption',
         'extranet',
         'firmware',
         'flexibility',
         'focus group',
         'forecast',
         'frame',
         'framework',
         'function',
         'functionalities',
         'Graphic Interface',
         'groupware',
         'Graphical User Interface',
         'hardware',
         'help-desk',
         'hierarchy',
         'hub',
         'implementation',
         'info-mediaries',
         'infrastructure',
         'initiative',
         'installation',
         'instruction set',
         'interface',
         'Internet solution',
         'intranet',
         'knowledge user',
         'knowledgebase',
         'Local Area Network',
         'leverage',
         'matrices',
         'matrix',
         'methodology',
         'middleware',
         'migration',
         'model',
         'moderator',
         'monitoring',
         'moratorium',
         'neural-net',
         'open architecture',
         'open system',
         'orchestration',
         'paradigm',
         'parallelism',
         'policy',
         'portal',
         'pricing structure',
         'process improvement',
         'product',
         'productivity',
         'project',
         'projection',
         'protocol',
         'secured line',
         'service-desk',
         'software',
         'solution',
         'standardization',
         'strategy',
         'structure',
         'success',
         'superstructure',
         'support',
         'synergy',
         'system engine',
         'task-force',
         'throughput',
         'time-frame',
         'toolset',
         'utilization',
         'website',
         'workforce'))

    bsWords = (
        ('implement',
         'utilize',
         'integrate',
         'streamline',
         'optimize',
         'evolve',
         'transform',
         'embrace',
         'enable',
         'orchestrate',
         'leverage',
         'reinvent',
         'aggregate',
         'architect',
         'enhance',
         'incentivize',
         'morph',
         'empower',
         'envisioneer',
         'monetize',
         'harness',
         'facilitate',
         'seize',
         'disintermediate',
         'synergize',
         'strategize',
         'deploy',
         'brand',
         'grow',
         'target',
         'syndicate',
         'synthesize',
         'deliver',
         'mesh',
         'incubate',
         'engage',
         'maximize',
         'benchmark',
         'expedite',
         're-intermediate',
         'whiteboard',
         'visualize',
         'repurpose',
         'innovate',
         'scale',
         'unleash',
         'drive',
         'extend',
         'engineer',
         'revolutionize',
         'generate',
         'exploit',
         'transition',
         'e-enable',
         'iterate',
         'cultivate',
         'matrix',
         'productize',
         'redefine',
         're-contextualize'),
        ('clicks-and-mortar',
         'value-added',
         'vertical',
         'proactive',
         'robust',
         'revolutionary',
         'scalable',
         'leading-edge',
         'innovative',
         'intuitive',
         'strategic',
         'e-business',
         'mission-critical',
         'sticky',
         'one-to-one',
         '24/7',
         'end-to-end',
         'global',
         'B2B',
         'B2C',
         'granular',
         'frictionless',
         'virtual',
         'viral',
         'dynamic',
         '24/365',
         'best-of-breed',
         'killer',
         'magnetic',
         'bleeding-edge',
         'web-enabled',
         'interactive',
         'dot-com',
         'back-end',
         'real-time',
         'efficient',
         'front-end',
         'distributed',
         'seamless',
         'extensible',
         'turn-key',
         'world-class',
         'open-source',
         'cross-platform',
         'cross-media',
         'synergistic',
         'bricks-and-clicks',
         'out-of-the-box',
         'enterprise',
         'integrated',
         'impactful',
         'wireless',
         'transparent',
         'next-generation',
         'cutting-edge',
         'user-centric',
         'visionary',
         'customized',
         'ubiquitous',
         'plug-and-play',
         'collaborative',
         'compelling',
         'holistic',
         'rich'),
        ('synergies',
         'web-readiness',
         'paradigms',
         'markets',
         'partnerships',
         'infrastructures',
         'platforms',
         'initiatives',
         'channels',
         'eyeballs',
         'communities',
         'ROI',
         'solutions',
         'e-tailers',
         'e-services',
         'action-items',
         'portals',
         'niches',
         'technologies',
         'content',
         'vortals',
         'supply-chains',
         'convergence',
         'relationships',
         'architectures',
         'interfaces',
         'e-markets',
         'e-commerce',
         'systems',
         'bandwidth',
         'info-mediaries',
         'models',
         'mindshare',
         'deliverables',
         'users',
         'schemas',
         'networks',
         'applications',
         'metrics',
         'e-business',
         'functionalities',
         'experiences',
         'web services',
         'methodologies'))

    def company(self):
        """
        :example 'Acme Ltd'
        """
        pattern = self.random_element(self.formats)
        return self.generator.parse(pattern)

    def company_suffix(self):
        """
        :example 'Ltd'
        """
        return self.random_element(self.company_suffixes)

    def catch_phrase(self):
        """
        :example 'Robust full-range hub'
        """
        result = []
        for word_list in self.catch_phrase_words:
            result.append(self.random_element(word_list))

        return " ".join(result)

    def bs(self):
        """
        :example 'integrate extensible convergence'
        """
        result = []
        for word_list in self.bsWords:
            result.append(self.random_element(word_list))

        return " ".join(result)



import os

TESTS_DIR = os.path.abspath(os.path.dirname(__file__))


TEST_APPS = [
    file.name
    for file in os.scandir(TESTS_DIR)
    if os.path.isdir(os.path.join(TESTS_DIR, file.name))
    and "." not in file.name
    and not file.name.startswith("__")
]


DJANGO_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.humanize",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
]

SCARLET_AND_OTHER_APPS = [
    "scarlet.cms",
    "scarlet.assets",
    "scarlet.accounts",
    "scarlet.versioning",
    "scarlet.scheduling",
    "scarlet.pagebuilder",
    "taggit",
    "webpack_loader",
]

urls = "cms_bundles.urls"


SECRET_KEY = "Please do not spew DeprecationWarnings"
SITE_ID = 1
INSTALLED_APPS = DJANGO_APPS + SCARLET_AND_OTHER_APPS + TEST_APPS
STATIC_URL = "static/"
ROOT_URLCONF = urls
USE_TZ = True
DATABASES = {
    "default": {
        "ENGINE": "scarlet.versioning.postgres_backend",
        "NAME": "scarlet",
        "USER": "scarlet",
        "PASSWORD": "scarlet",
        "HOST": "localhost",
        "PORT": "",
    },
}
MIDDLEWARE = (
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
)
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ],
            "debug": True,
        },
    },
]
